from abc import ABC
from dataclasses import field
from typing import ClassVar, Dict, Optional, Tuple, Type

import bitsandbytes as bnb
import torch
from marshmallow import fields, ValidationError

import ludwig.schema.utils as schema_utils
from ludwig.api_annotations import DeveloperAPI
from ludwig.schema.metadata import OPTIMIZER_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json, ParameterMetadata
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils.registry import Registry

optimizer_registry = Registry()


@DeveloperAPI
def register_optimizer(name: str):
    def wrap(optimizer_config: BaseOptimizerConfig):
        optimizer_registry[name] = (optimizer_config.optimizer_class, optimizer_config)
        return optimizer_config

    return wrap


@DeveloperAPI
def get_optimizer_cls(name: str):
    """Get the optimizer schema class from the optimizer schema class registry."""
    return optimizer_registry[name][1]


@DeveloperAPI
@ludwig_dataclass
class BaseOptimizerConfig(schema_utils.BaseMarshmallowConfig, ABC):
    """Base class for optimizers. Not meant to be used directly.

    The dataclass format prevents arbitrary properties from being set. Consequently, in child classes, all properties
    from the corresponding `torch.optim.Optimizer` class are copied over: check each class to check which attributes are
    different from the torch-specified defaults.
    """

    optimizer_class: ClassVar[Optional[torch.optim.Optimizer]] = None
    "Class variable pointing to the corresponding `torch.optim.Optimizer` class."

    type: str
    """Name corresponding to an optimizer `ludwig.modules.optimization_modules.optimizer_registry`.

    Technically mutable, but attempting to load a derived optimizer with `type` set to a mismatched value will result in
    a `ValidationError`.
    """

    @property
    def is_paged(self) -> bool:
        """Returns True if the optimizer is a Paged optimizer."""
        return False

    @property
    def is_8bit(self) -> bool:
        """Returns True if the optimizer is an 8-bit optimizer."""
        return False


@DeveloperAPI
@register_optimizer(name="sgd")
@ludwig_dataclass
class SGDOptimizerConfig(BaseOptimizerConfig):
    """Parameters for stochastic gradient descent."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.SGD
    """Points to `torch.optim.SGD`."""

    type: str = schema_utils.ProtectedString("sgd")
    """Must be 'sgd' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry` (default:
       'sgd')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.SGD.html#torch.optim.SGD :
    momentum: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Momentum factor.",
        parameter_metadata=OPTIMIZER_METADATA["momentum"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Weight decay ($L2$ penalty).",
        parameter_metadata=OPTIMIZER_METADATA["weight_decay"],
    )

    dampening: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Dampening for momentum.",
        parameter_metadata=OPTIMIZER_METADATA["dampening"],
    )

    nesterov: bool = schema_utils.Boolean(
        default=False,
        description="Enables Nesterov momentum.",
        parameter_metadata=OPTIMIZER_METADATA["nesterov"],
    )


@DeveloperAPI
@register_optimizer(name="sgd_8bit")
@ludwig_dataclass
class SGD8BitOptimizerConfig(SGDOptimizerConfig):
    """Parameters for stochastic gradient descent."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.SGD8bit

    type: str = schema_utils.ProtectedString("sgd_8bit")

    block_wise: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use block wise update.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="lbfgs")
@ludwig_dataclass
class LBFGSOptimizerConfig(BaseOptimizerConfig):
    """Parameters for stochastic gradient descent."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.LBFGS
    """Points to `torch.optim.LBFGS`."""

    type: str = schema_utils.ProtectedString("lbfgs")
    """Must be 'lbfgs' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry` (default:
       'lbfgs')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.LBFGS.html#torch.optim.LBFGS
    max_iter: int = schema_utils.Integer(
        default=20,
        description="Maximum number of iterations per optimization step.",
        parameter_metadata=OPTIMIZER_METADATA["max_iter"],
    )

    max_eval: int = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="Maximum number of function evaluations per optimization step. Default: `max_iter` * 1.25.",
        parameter_metadata=OPTIMIZER_METADATA["max_eval"],
    )

    tolerance_grad: float = schema_utils.NonNegativeFloat(
        default=1e-07,
        description="Termination tolerance on first order optimality.",
        parameter_metadata=OPTIMIZER_METADATA["tolerance_grad"],
    )

    tolerance_change: float = schema_utils.NonNegativeFloat(
        default=1e-09,
        description="Termination tolerance on function value/parameter changes.",
        parameter_metadata=OPTIMIZER_METADATA["tolerance_change"],
    )

    history_size: int = schema_utils.Integer(
        default=100, description="Update history size.", parameter_metadata=OPTIMIZER_METADATA["history_size"]
    )

    line_search_fn: str = schema_utils.StringOptions(
        ["strong_wolfe"],
        default=None,
        allow_none=True,
        description="Line search function to use.",
        parameter_metadata=OPTIMIZER_METADATA["line_search_fn"],
    )


@DeveloperAPI
@register_optimizer(name="adam")
@ludwig_dataclass
class AdamOptimizerConfig(BaseOptimizerConfig):
    """Parameters for adam optimization."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.Adam
    """Points to `torch.optim.Adam`."""

    type: str = schema_utils.ProtectedString("adam")
    """Must be 'adam' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'adam')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam :
    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0, description="Weight decay (L2 penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )

    amsgrad: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use the AMSGrad variant of this algorithm from the paper 'On the Convergence of Adam "
        "and Beyond'.",
        parameter_metadata=OPTIMIZER_METADATA["amsgrad"],
    )


@DeveloperAPI
@register_optimizer(name="adam_8bit")
@ludwig_dataclass
class Adam8BitOptimizerConfig(AdamOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.Adam8bit

    type: str = schema_utils.ProtectedString("adam_8bit")

    block_wise: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use block wise update.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="paged_adam")
@ludwig_dataclass
class PagedAdamOptimizerConfig(Adam8BitOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedAdam

    type: str = schema_utils.ProtectedString("paged_adam")

    @property
    def is_paged(self) -> bool:
        return True

    @property
    def is_8bit(self) -> bool:
        return False


@DeveloperAPI
@register_optimizer(name="paged_adam_8bit")
@ludwig_dataclass
class PagedAdam8BitOptimizerConfig(PagedAdamOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedAdam8bit

    type: str = schema_utils.ProtectedString("paged_adam_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="adamw")
@ludwig_dataclass
class AdamWOptimizerConfig(BaseOptimizerConfig):
    """Parameters for adamw optimization."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.AdamW
    """Points to `torch.optim.AdamW`."""

    type: str = schema_utils.ProtectedString("adamw")
    """Must be 'adamw' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'adamw')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam :
    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0, description="Weight decay ($L2$ penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )

    amsgrad: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use the AMSGrad variant of this algorithm from the paper 'On the Convergence of Adam "
        "and Beyond'. ",
        parameter_metadata=OPTIMIZER_METADATA["amsgrad"],
    )


@DeveloperAPI
@register_optimizer(name="adamw_8bit")
@ludwig_dataclass
class AdamW8BitOptimizerConfig(AdamWOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.AdamW8bit

    type: str = schema_utils.ProtectedString("adamw_8bit")

    block_wise: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use block wise update.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="paged_adamw")
@ludwig_dataclass
class PagedAdamWOptimizerConfig(AdamW8BitOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedAdamW

    type: str = schema_utils.ProtectedString("paged_adamw")

    @property
    def is_paged(self) -> bool:
        return True

    @property
    def is_8bit(self) -> bool:
        return False


@DeveloperAPI
@register_optimizer(name="paged_adamw_8bit")
@ludwig_dataclass
class PagedAdamW8BitOptimizerConfig(PagedAdamWOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedAdamW8bit

    type: str = schema_utils.ProtectedString("paged_adamw_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="adadelta")
@ludwig_dataclass
class AdadeltaOptimizerConfig(BaseOptimizerConfig):
    """Parameters for adadelta optimization."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.Adadelta
    """Points to `torch.optim.Adadelta`."""

    type: str = schema_utils.ProtectedString("adadelta")
    """Must be 'adadelta' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'adadelta')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.Adadelta.html#torch.optim.Adadelta :
    rho: float = schema_utils.FloatRange(
        default=0.9,
        min=0,
        max=1,
        description="Coefficient used for computing a running average of squared gradients.",
        parameter_metadata=OPTIMIZER_METADATA["rho"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-06,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0, description="Weight decay ($L2$ penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )


@DeveloperAPI
@register_optimizer(name="adagrad")
@ludwig_dataclass
class AdagradOptimizerConfig(BaseOptimizerConfig):
    """Parameters for adagrad optimization."""

    # Example docstring
    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.Adagrad
    """Points to `torch.optim.Adagrad`."""

    type: str = schema_utils.ProtectedString("adagrad")
    """Must be 'adagrad' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'adagrad')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.Adagrad.html#torch.optim.Adagrad :
    initial_accumulator_value: float = schema_utils.NonNegativeFloat(
        default=0, description="", parameter_metadata=OPTIMIZER_METADATA["initial_accumulator_value"]
    )

    lr_decay: float = schema_utils.FloatRange(
        default=0, description="Learning rate decay.", parameter_metadata=OPTIMIZER_METADATA["lr_decay"]
    )

    weight_decay: float = schema_utils.FloatRange(
        default=0, description="Weight decay ($L2$ penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )

    eps: float = schema_utils.FloatRange(
        default=1e-10,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )


@DeveloperAPI
@register_optimizer(name="adagrad_8bit")
@ludwig_dataclass
class Adagrad8BitOptimizerConfig(AdagradOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.Adagrad8bit

    type: str = schema_utils.ProtectedString("adagrad_8bit")

    block_wise: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use block wise update.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="adamax")
@ludwig_dataclass
class AdamaxOptimizerConfig(BaseOptimizerConfig):
    """Parameters for adamax optimization."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.Adamax
    """Points to `torch.optim.Adamax`."""

    type: str = schema_utils.ProtectedString("adamax")
    """Must be 'adamax' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'adamax')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.Adamax.html#torch.optim.Adamax :
    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0, description="Weight decay ($L2$ penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )


# NOTE: keep ftrl and nadam optimizers out of registry:
# @register_optimizer(name="ftrl")
@DeveloperAPI
@ludwig_dataclass
class FtrlOptimizerConfig(BaseOptimizerConfig):
    # optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.Ftrl
    type: str = schema_utils.ProtectedString("ftrl")

    learning_rate_power: float = schema_utils.FloatRange(
        default=-0.5, max=0, parameter_metadata=OPTIMIZER_METADATA["learning_rate_power"]
    )

    initial_accumulator_value: float = schema_utils.NonNegativeFloat(
        default=0.1, parameter_metadata=OPTIMIZER_METADATA["initial_accumulator_value"]
    )

    l1_regularization_strength: float = schema_utils.NonNegativeFloat(
        default=0.0, parameter_metadata=OPTIMIZER_METADATA["l1_regularization_strength"]
    )

    l2_regularization_strength: float = schema_utils.NonNegativeFloat(
        default=0.0, parameter_metadata=OPTIMIZER_METADATA["l2_regularization_strength"]
    )


@DeveloperAPI
@register_optimizer(name="nadam")
@ludwig_dataclass
class NadamOptimizerConfig(BaseOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.NAdam
    """Points to `torch.optim.NAdam`."""

    type: str = schema_utils.ProtectedString("nadam")

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.NAdam.html#torch.optim.NAdam :

    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0, description="Weight decay ($L2$ penalty).", parameter_metadata=OPTIMIZER_METADATA["weight_decay"]
    )

    momentum_decay: float = schema_utils.NonNegativeFloat(
        default=4e-3, description="Momentum decay.", parameter_metadata=OPTIMIZER_METADATA["momentum_decay"]
    )


@DeveloperAPI
@register_optimizer(name="rmsprop")
@ludwig_dataclass
class RMSPropOptimizerConfig(BaseOptimizerConfig):
    """Parameters for rmsprop optimization."""

    optimizer_class: ClassVar[torch.optim.Optimizer] = torch.optim.RMSprop
    """Points to `torch.optim.RMSprop`."""

    type: str = schema_utils.ProtectedString("rmsprop")
    """Must be 'rmsprop' - corresponds to name in `ludwig.modules.optimization_modules.optimizer_registry`
       (default: 'rmsprop')"""

    # Defaults taken from https://pytorch.org/docs/stable/generated/torch.optim.RMSprop.html#torch.optim.RMSprop:
    momentum: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Momentum factor.",
        parameter_metadata=OPTIMIZER_METADATA["momentum"],
    )

    alpha: float = schema_utils.NonNegativeFloat(
        default=0.99,
        description="Smoothing constant.",
        parameter_metadata=OPTIMIZER_METADATA["alpha"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    centered: bool = schema_utils.Boolean(
        default=False,
        description="If True, computes the centered RMSProp, and the gradient is normalized by an estimation of its "
        "variance.",
        parameter_metadata=OPTIMIZER_METADATA["centered"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(default=0.0, description="Weight decay ($L2$ penalty).")


@DeveloperAPI
@register_optimizer(name="rmsprop_8bit")
@ludwig_dataclass
class RMSProp8BitOptimizerConfig(RMSPropOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.RMSprop8bit

    type: str = schema_utils.ProtectedString("rmsprop_8bit")

    block_wise: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use block wise update.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="lamb")
@ludwig_dataclass
class LAMBOptimizerConfig(BaseOptimizerConfig):
    """Layer-wise Adaptive Moments optimizer for Batch training.

    Paper: https://arxiv.org/pdf/1904.00962.pdf
    """

    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.LAMB

    type: str = schema_utils.ProtectedString("lamb")

    bias_correction: bool = schema_utils.Boolean(
        default=True,
    )

    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    eps: float = schema_utils.NonNegativeFloat(
        default=1e-08,
        description="Term added to the denominator to improve numerical stability.",
        parameter_metadata=OPTIMIZER_METADATA["eps"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Weight decay (L2 penalty).",
        parameter_metadata=OPTIMIZER_METADATA["weight_decay"],
    )

    amsgrad: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use the AMSGrad variant of this algorithm from the paper 'On the Convergence of Adam "
        "and Beyond'.",
        parameter_metadata=OPTIMIZER_METADATA["amsgrad"],
    )

    adam_w_mode: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use the AdamW mode of this algorithm from the paper "
        "'Decoupled Weight Decay Regularization'.",
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    block_wise: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use block wise update.",
    )

    max_unorm: float = schema_utils.FloatRange(
        default=1.0,
        min=0.0,
        max=1.0,
    )


@DeveloperAPI
@register_optimizer(name="lamb_8bit")
@ludwig_dataclass
class LAMB8BitOptimizerConfig(LAMBOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.LAMB8bit

    type: str = schema_utils.ProtectedString("lamb_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="lars")
@ludwig_dataclass
class LARSOptimizerConfig(BaseOptimizerConfig):
    """Layerwise Adaptive Rate Scaling.

    Paper: https://arxiv.org/pdf/1708.03888.pdf
    """

    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.LARS

    type: str = schema_utils.ProtectedString("lars")

    # 0.9 taken from the original paper - momentum requires a non zero value
    # https://arxiv.org/pdf/1708.03888v3.pdf
    momentum: float = schema_utils.FloatRange(
        default=0.9,
        min=0.0,
        max=1.0,
        min_inclusive=False,
        description="Momentum factor.",
        parameter_metadata=OPTIMIZER_METADATA["momentum"],
    )

    dampening: float = schema_utils.FloatRange(
        default=0.0,
        min=0.0,
        max=1.0,
        description="Dampening for momentum.",
        parameter_metadata=OPTIMIZER_METADATA["dampening"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Weight decay (L2 penalty).",
        parameter_metadata=OPTIMIZER_METADATA["weight_decay"],
    )

    nesterov: bool = schema_utils.Boolean(
        default=False,
        description="Enables Nesterov momentum.",
        parameter_metadata=OPTIMIZER_METADATA["nesterov"],
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    max_unorm: float = schema_utils.FloatRange(
        default=1.0,
        min=0.0,
        max=1.0,
    )


@DeveloperAPI
@register_optimizer(name="lars_8bit")
@ludwig_dataclass
class LARS8BitOptimizerConfig(LARSOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.LARS8bit

    type: str = schema_utils.ProtectedString("lars_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="lion")
@ludwig_dataclass
class LIONOptimizerConfig(BaseOptimizerConfig):
    """Evolved Sign Momentum.

    Paper: https://arxiv.org/pdf/2302.06675.pdf
    """

    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.Lion

    type: str = schema_utils.ProtectedString("lion")

    betas: Tuple[float, float] = schema_utils.FloatRangeTupleDataclassField(
        default=(0.9, 0.999),
        description="Coefficients used for computing running averages of gradient and its square.",
        parameter_metadata=OPTIMIZER_METADATA["betas"],
    )

    weight_decay: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Weight decay (L2 penalty).",
        parameter_metadata=OPTIMIZER_METADATA["weight_decay"],
    )

    percentile_clipping: int = schema_utils.IntegerRange(
        default=100,
        min=0,
        max=100,
        description="Percentile clipping.",
    )

    block_wise: bool = schema_utils.Boolean(
        default=True,
        description="Whether to use block wise update.",
    )


@DeveloperAPI
@register_optimizer(name="lion_8bit")
@ludwig_dataclass
class LION8BitOptimizerConfig(LIONOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.Lion8bit

    type: str = schema_utils.ProtectedString("lion_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="paged_lion")
@ludwig_dataclass
class PagedLionOptimizerConfig(LIONOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedLion

    type: str = schema_utils.ProtectedString("paged_lion")

    @property
    def is_paged(self) -> bool:
        return True


@DeveloperAPI
@register_optimizer(name="paged_lion_8bit")
@ludwig_dataclass
class PagedLion8BitOptimizerConfig(PagedLionOptimizerConfig):
    optimizer_class: ClassVar[torch.optim.Optimizer] = bnb.optim.PagedLion8bit

    type: str = schema_utils.ProtectedString("paged_lion_8bit")

    @property
    def is_8bit(self) -> bool:
        return True


@DeveloperAPI
def get_optimizer_conds():
    """Returns a JSON schema of conditionals to validate against optimizer types defined in
    `ludwig.modules.optimization_modules.optimizer_registry`."""
    conds = []
    for optimizer in optimizer_registry:
        optimizer_cls = optimizer_registry[optimizer][1]
        other_props = schema_utils.unload_jsonschema_from_marshmallow_class(optimizer_cls)["properties"]
        schema_utils.remove_duplicate_fields(other_props)
        preproc_cond = schema_utils.create_cond(
            {"type": optimizer},
            other_props,
        )
        conds.append(preproc_cond)
    return conds


@DeveloperAPI
def OptimizerDataclassField(default="adam", description="", parameter_metadata: ParameterMetadata = None):
    """Custom dataclass field that when used inside of a dataclass will allow any optimizer in
    `ludwig.modules.optimization_modules.optimizer_registry`.

    Sets default optimizer to 'adam'.

    :param default: Dict specifying an optimizer with a `type` field and its associated parameters. Will attempt to use
           `type` to load optimizer from registry with given params. (default: {"type": "adam"}).
    :return: Initialized dataclass field that converts untyped dicts with params to optimizer dataclass instances.
    """

    class OptimizerSelection(schema_utils.TypeSelection):
        """Custom marshmallow field that deserializes a dict to a valid optimizer from
        `ludwig.modules.optimization_modules.optimizer_registry` and creates a corresponding `oneOf` JSON schema
        for external usage."""

        def __init__(self):
            super().__init__(
                registry=optimizer_registry,
                default_value=default,
                description=description,
                parameter_metadata=parameter_metadata,
            )

        def get_schema_from_registry(self, key: str) -> Type[schema_utils.BaseMarshmallowConfig]:
            return get_optimizer_cls(key)

        def _jsonschema_type_mapping(self):
            # Note that this uses the same conditional pattern as combiners:
            return {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": list(optimizer_registry.keys()),
                        "default": default,
                        "description": "The type of optimizer to use during the learning process",
                    },
                },
                "title": "optimizer_options",
                "allOf": get_optimizer_conds(),
                "required": ["type"],
                "description": description,
            }

    return OptimizerSelection().get_default_field()


@DeveloperAPI
@ludwig_dataclass
class GradientClippingConfig(schema_utils.BaseMarshmallowConfig):
    """Dataclass that holds gradient clipping parameters."""

    clipglobalnorm: Optional[float] = schema_utils.FloatRange(
        default=0.5,
        allow_none=True,
        description="Maximum allowed norm of the gradients",
        parameter_metadata=OPTIMIZER_METADATA["gradient_clipping"],
    )

    # TODO(travis): is this redundant with `clipglobalnorm`?
    clipnorm: Optional[float] = schema_utils.FloatRange(
        default=None,
        allow_none=True,
        description="Maximum allowed norm of the gradients",
        parameter_metadata=OPTIMIZER_METADATA["gradient_clipping"],
    )

    clipvalue: Optional[float] = schema_utils.FloatRange(
        default=None,
        allow_none=True,
        description="Maximum allowed value of the gradients",
        parameter_metadata=OPTIMIZER_METADATA["gradient_clipping"],
    )


@DeveloperAPI
def GradientClippingDataclassField(description: str, default: Dict = {}):
    """Returns custom dataclass field for `ludwig.modules.optimization_modules.GradientClippingConfig`. Allows
    `None` by default.

    :param description: Description of the gradient dataclass field
    :param default: dict that specifies clipping param values that will be loaded by its schema class (default: {}).
    """
    allow_none = True

    class GradientClippingMarshmallowField(fields.Field):
        """Custom marshmallow field class for gradient clipping.

        Deserializes a dict to a valid instance of `ludwig.modules.optimization_modules.GradientClippingConfig` and
        creates a corresponding JSON schema for external usage.
        """

        def _deserialize(self, value, attr, data, **kwargs):
            if value is None:
                return value
            if isinstance(value, dict):
                try:
                    return GradientClippingConfig.Schema().load(value)
                except (TypeError, ValidationError):
                    raise ValidationError(
                        f"Invalid params for gradient clipping: {value}, see GradientClippingConfig class."
                    )
            raise ValidationError("Field should be None or dict")

        def _jsonschema_type_mapping(self):
            return {
                "oneOf": [
                    {"type": "null", "title": "disabled", "description": "Disable gradient clipping."},
                    {
                        **schema_utils.unload_jsonschema_from_marshmallow_class(GradientClippingConfig),
                        "title": "enabled_options",
                    },
                ],
                "title": "gradient_clipping_options",
                "description": description,
            }

    if not isinstance(default, dict):
        raise ValidationError(f"Invalid default: `{default}`")

    def load_default():
        return GradientClippingConfig.Schema().load(default)

    dump_default = GradientClippingConfig.Schema().dump(default)

    return field(
        metadata={
            "marshmallow_field": GradientClippingMarshmallowField(
                allow_none=allow_none,
                load_default=load_default,
                dump_default=dump_default,
                metadata={
                    "description": description,
                    "parameter_metadata": convert_metadata_to_json(OPTIMIZER_METADATA["gradient_clipping"]),
                },
            )
        },
        default_factory=load_default,
    )



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import HUBER, MEAN_SQUARED_ERROR, MODEL_ECD, TIMESERIES, VECTOR
from ludwig.schema import utils as schema_utils
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import DecoderDataclassField
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig, BaseOutputFeatureConfig
from ludwig.schema.features.loss.loss import BaseLossConfig
from ludwig.schema.features.loss.utils import LossDataclassField
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import (
    ecd_defaults_config_registry,
    ecd_input_config_registry,
    ecd_output_config_registry,
    input_mixin_registry,
    output_mixin_registry,
)
from ludwig.schema.metadata import FEATURE_METADATA
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@input_mixin_registry.register(TIMESERIES)
@ludwig_dataclass
class TimeseriesInputFeatureConfigMixin(BaseMarshmallowConfig):
    """TimeseriesInputFeatureConfigMixin is a dataclass that configures the parameters used in both the timeseries
    input feature and the timeseries global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=TIMESERIES)

    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=TIMESERIES,
        default="parallel_cnn",
    )


@DeveloperAPI
@ecd_input_config_registry.register(TIMESERIES)
@ludwig_dataclass
class TimeseriesInputFeatureConfig(TimeseriesInputFeatureConfigMixin, BaseInputFeatureConfig):
    """TimeseriesInputFeatureConfig is a dataclass that configures the parameters used for a timeseries input
    feature."""

    type: str = schema_utils.ProtectedString(TIMESERIES)


@DeveloperAPI
@output_mixin_registry.register(TIMESERIES)
@ludwig_dataclass
class TimeseriesOutputFeatureConfigMixin(BaseMarshmallowConfig):
    """TimeseriesOutputFeatureConfigMixin configures the parameters used in both the timeseries output feature and
    the timeseries global defaults section of the Ludwig Config."""

    decoder: BaseDecoderConfig = DecoderDataclassField(
        MODEL_ECD,
        feature_type=TIMESERIES,
        default="projector",
    )

    loss: BaseLossConfig = LossDataclassField(
        feature_type=TIMESERIES,
        default=HUBER,
    )


@DeveloperAPI
@ecd_output_config_registry.register(TIMESERIES)
@ludwig_dataclass
class TimeseriesOutputFeatureConfig(BaseOutputFeatureConfig, TimeseriesOutputFeatureConfigMixin):
    """TimeseriesOutputFeatureConfig configures the parameters used for a timeseries output feature."""

    type: str = schema_utils.ProtectedString(TIMESERIES)

    dependencies: list = schema_utils.List(
        default=[],
        description="List of input features that this feature depends on.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["dependencies"],
    )

    default_validation_metric: str = schema_utils.StringOptions(
        [MEAN_SQUARED_ERROR],
        default=MEAN_SQUARED_ERROR,
        description="Internal parameter.",
        parameter_metadata=INTERNAL_ONLY,
    )

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type="timeseries_output")

    reduce_dependencies: str = schema_utils.ReductionOptions(
        default=None,
        description="How to reduce the dependencies of the output feature.",
        parameter_metadata=FEATURE_METADATA[VECTOR]["reduce_dependencies"],
    )

    reduce_input: str = schema_utils.ReductionOptions(
        default=None,
        description="How to reduce an input that is not a vector, but a matrix or a higher order tensor, on the first "
        "dimension (second if you count the batch dimension)",
        parameter_metadata=FEATURE_METADATA[VECTOR]["reduce_input"],
    )

    horizon: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Internal parameter. Obtained from preprocessing",
        parameter_metadata=INTERNAL_ONLY,
    )


@DeveloperAPI
@ecd_defaults_config_registry.register(TIMESERIES)
@ludwig_dataclass
class TimeseriesDefaultsConfig(TimeseriesInputFeatureConfigMixin, TimeseriesOutputFeatureConfigMixin):
    pass



import warnings

from transformers import BitsAndBytesConfig

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import LLM_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json
from ludwig.schema.utils import ludwig_dataclass

warnings.filterwarnings(
    action="ignore",
    category=UserWarning,
    module="bitsandbytes.cuda_setup.main",
)


@DeveloperAPI
@ludwig_dataclass
class QuantizationConfig(schema_utils.BaseMarshmallowConfig):
    bits: int = schema_utils.IntegerOptions(
        options=[4, 8],
        default=4,
        description="The quantization level to apply to weights on load.",
        parameter_metadata=LLM_METADATA["quantization"]["bits"],
    )

    llm_int8_threshold: float = schema_utils.NonNegativeFloat(
        default=6.0,
        description=(
            "This corresponds to the outlier threshold for outlier detection as described in `LLM.int8() : 8-bit "
            "Matrix Multiplication for Transformers at Scale` paper: https://arxiv.org/abs/2208.07339. Any hidden "
            "states value that is above this threshold will be considered an outlier and the operation on those "
            "values will be done in fp16. Values are usually normally distributed, that is, most values are in the "
            "range [-3.5, 3.5], but there are some exceptional systematic outliers that are very differently "
            "distributed for large models. These outliers are often in the interval [-60, -6] or [6, 60]. Int8 "
            "quantization works well for values of magnitude ~5, but beyond that, there is a significant performance "
            "penalty. A good default threshold is 6, but a lower threshold might be needed for more unstable models "
            "(small models, fine-tuning)."
        ),
    )

    llm_int8_has_fp16_weight: bool = schema_utils.Boolean(
        default=False,
        description=(
            "This flag runs LLM.int8() with 16-bit main weights. This is useful for fine-tuning as the weights do "
            "not have to be converted back and forth for the backward pass."
        ),
    )

    bnb_4bit_compute_dtype: str = schema_utils.StringOptions(
        options=["float32", "float16", "bfloat16"],
        default="float16",
        description=(
            "This sets the computational type which might be different than the input type. For example, inputs "
            "might be fp32, but computation can be set to bf16 for speedups."
        ),
    )

    bnb_4bit_use_double_quant: bool = schema_utils.Boolean(
        default=True,
        description=(
            "This flag is used for nested quantization where the quantization constants from the first quantization "
            "are quantized again."
        ),
    )

    bnb_4bit_quant_type: str = schema_utils.StringOptions(
        options=["fp4", "nf4"],
        default="nf4",
        description="This sets the quantization data type in the bnb.nn.Linear4Bit layers.",
    )

    def to_bitsandbytes(self) -> BitsAndBytesConfig:
        return BitsAndBytesConfig(
            load_in_4bit=self.bits == 4,
            load_in_8bit=self.bits == 8,
            llm_int8_threshold=self.llm_int8_threshold,
            llm_int8_has_fp16_weight=self.llm_int8_has_fp16_weight,
            bnb_4bit_compute_dtype=self.bnb_4bit_compute_dtype,
            bnb_4bit_use_double_quant=self.bnb_4bit_use_double_quant,
            bnb_4bit_quant_type=self.bnb_4bit_quant_type,
        )


@DeveloperAPI
class QuantizationConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(QuantizationConfig, default_missing=True)

    def _jsonschema_type_mapping(self):
        return {
            "oneOf": [
                {
                    "type": "null",
                    "title": "disabled",
                    "description": "Disable quantization.",
                    "parameter_metadata": convert_metadata_to_json(LLM_METADATA["quantization"]["_oneOf"]["none"]),
                },
                {
                    **schema_utils.unload_jsonschema_from_marshmallow_class(QuantizationConfig),
                    "title": "enabled",
                    "description": "Set quantization options.",
                    "parameter_metadata": convert_metadata_to_json(LLM_METADATA["quantization"]["_oneOf"]["object"]),
                },
            ],
            "title": "quantization",
            "description": "Set quantization options.",
            "parameter_metadata": convert_metadata_to_json(LLM_METADATA["quantization"]["_meta"]),
        }



import logging
import os
from abc import ABC, abstractmethod
from typing import Optional

from huggingface_hub import HfApi, login

logger = logging.getLogger(__name__)


class BaseModelUpload(ABC):
    """Abstract base class for uploading trained model artifacts to different repositories.

    This class defines the interface for uploading trained model artifacts to various repositories such as Huggingface
    Hub, without specifying the concrete implementation for each repository. Subclasses of this base class must
    implement the 'login' and 'upload' methods.
    """

    @abstractmethod
    def login(self):
        """Abstract method to handle authentication with the target repository.

        Subclasses must implement this method to provide the necessary authentication
        mechanisms required by the repository where the model artifacts will be uploaded.

        Raises:
            NotImplementedError: If this method is not implemented in the subclass.
        """
        raise NotImplementedError()

    @abstractmethod
    def upload(
        self,
        repo_id: str,
        model_path: str,
        repo_type: Optional[str] = None,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
        commit_description: Optional[str] = None,
    ) -> bool:
        """Abstract method to upload trained model artifacts to the target repository.

        Subclasses must implement this method to define the process of pushing model
        artifacts to the respective repository. This may include creating a new model version,
        uploading model files, and any other specific steps required by the model repository
        service.

        Returns:
            bool: True if the model artifacts were successfully uploaded, False otherwise.

        Raises:
            NotImplementedError: If this method is not implemented in the subclass.
        """
        raise NotImplementedError()

    @staticmethod
    def _validate_upload_parameters(
        repo_id: str,
        model_path: str,
        repo_type: Optional[str] = None,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
        commit_description: Optional[str] = None,
    ):
        """Validate parameters before uploading trained model artifacts.

        This method checks if the input parameters meet the necessary requirements before uploading
        trained model artifacts to the target repository.

        Args:
            repo_id (str): The ID of the target repository. It must be a namespace (user or an organization)
                and a repository name separated by a '/'. For example, if your HF username is 'johndoe' and you
                want to create a repository called 'test', the repo_id should be 'johndoe/test'.
            model_path (str): The path to the directory containing the trained model artifacts. It should contain
                the model's weights, usually saved under 'model/model_weights'.
            repo_type (str, optional): The type of the repository. Not used in the base class, but subclasses
                may use it for specific repository implementations. Defaults to None.
            private (bool, optional): Whether the repository should be private or not. Not used in the base class,
                but subclasses may use it for specific repository implementations. Defaults to False.
            commit_message (str, optional): A message to attach to the commit when uploading to version control
                systems. Not used in the base class, but subclasses may use it for specific repository
                implementations. Defaults to None.
            commit_description (str, optional): A description of the commit when uploading to version control
                systems. Not used in the base class, but subclasses may use it for specific repository
                implementations. Defaults to None.

        Raises:
            AssertionError: If the repo_id does not have both a namespace and a repo name separated by a '/'.
            FileNotFoundError: If the model_path does not exist.
            Exception: If the trained model artifacts are not found at the expected location within model_path, or
                if the artifacts are not in the required format (i.e., 'pytorch_model.bin' or 'adapter_model.bin').
        """
        # Validate repo_id has both a namespace and a repo name
        assert "/" in repo_id, (
            "`repo_id` must be a namespace (user or an organization) and a repo name separated by a `/`."
            " For example, if your HF username is `johndoe` and you want to create a repository called `test`, the"
            " repo_id should be johndoe/test"
        )

        # Make sure the model's save path is actually a valid path
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"The path '{model_path}' does not exist.")

        # Make sure the model is actually trained
        trained_model_artifacts_path = os.path.join(model_path, "model", "model_weights")
        if not os.path.exists(trained_model_artifacts_path):
            raise Exception(
                f"Model artifacts not found at {trained_model_artifacts_path}. "
                f"It is possible that model at '{model_path}' hasn't been trained yet, or something went"
                "wrong during training where the model's weights were not saved."
            )

        # Make sure the model's saved artifacts either contain:
        # 1. pytorch_model.bin -> regular model training, such as ECD or for LLMs
        # 2. adapter_model.bin -> LLM fine-tuning using PEFT
        files = set(os.listdir(trained_model_artifacts_path))
        if "pytorch_model.bin" not in files and "adapter_model.bin" not in files:
            raise Exception(
                f"Can't find model weights at {trained_model_artifacts_path}. Trained model weights should "
                "either be saved as `pytorch_model.bin` for regular model training, or have `adapter_model.bin`"
                "if using parameter efficient fine-tuning methods like LoRA."
            )


class HuggingFaceHub(BaseModelUpload):
    def __init__(self):
        self.api = None

    def login(self):
        """Login to huggingface hub using the token stored in ~/.cache/huggingface/token and returns a HfApi client
        object that can be used to interact with HF Hub."""
        cached_token_path = os.path.join(os.path.expanduser("~"), ".cache", "huggingface", "token")

        if not os.path.exists(cached_token_path):
            login(add_to_git_credential=True)

        with open(cached_token_path) as f:
            hf_token = f.read()

        hf_api = HfApi(token=hf_token)
        assert hf_api.token == hf_token

        self.api = hf_api

    def upload(
        self,
        repo_id: str,
        model_path: str,
        repo_type: Optional[str] = None,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
        commit_description: Optional[str] = None,
    ) -> bool:
        """Create an empty repo on the HuggingFace Hub and upload trained model artifacts to that repo.

        Args:
            repo_id (`str`):
                A namespace (user or an organization) and a repo name separated
                by a `/`.
            model_path (`str`):
                The path of the saved model. This is the top level directory where
                the models weights as well as other associated training artifacts
                are saved.
            repo_type (`str`, *optional*):
                Set to `"dataset"` or `"space"` if uploading to a dataset or
                space, `None` or `"model"` if uploading to a model. Default is
                `None`.
            private (`bool`, *optional*, defaults to `False`):
                Whether the model repo should be private.
            commit_message (`str`, *optional*):
                The summary / title / first line of the generated commit. Defaults to:
                `f"Upload {path_in_repo} with huggingface_hub"`
            commit_description (`str` *optional*):
                The description of the generated commit
        """
        # Validate upload parameters are in the right format
        HuggingFaceHub._validate_upload_parameters(
            repo_id,
            model_path,
            repo_type,
            private,
            commit_message,
            commit_description,
        )

        # Create empty model repo using repo_id, but it is okay if it already exists.
        self.api.create_repo(
            repo_id=repo_id,
            private=private,
            repo_type=repo_type,
            exist_ok=True,
        )

        # Upload all artifacts in model weights folder
        upload_path = self.api.upload_folder(
            folder_path=os.path.join(model_path, "model", "model_weights"),
            repo_id=repo_id,
            repo_type=repo_type,
            commit_message=commit_message,
            commit_description=commit_description,
        )

        if upload_path:
            logger.info(f"Model uploaded to `{upload_path}` with repository name `{repo_id}`")
            return True

        return False



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from dataclasses_json import dataclass_json

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.misc_utils import memoized_method


@DeveloperAPI
class ExpectedImpact(int, Enum):
    """The expected impact of determining a "good" value for a specific parameter.

    - HIGH: this parameter should almost always be included in a hyperopt run and can make or break a good model.
    - MEDIUM: this parameter can sometimes make or break a good model.
    - LOW: this parameter usually does not have a significant impact on model performance.
    """

    UNKNOWN = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


@DeveloperAPI
class ComputeTier(int, Enum):
    """The compute tier defines the type of compute resources that a model typically needs to get good
    throughput."""

    CPU = 0
    """Model can train effectively on CPU hardware."""

    GPU_LOW = 1
    """Model can train effectively on commodity GPU hardware, or inference optimized SKUs like NVIDIA T4."""

    GPU_MEDIUM = 2
    """Model can train effectively on training-optimized GPU hardware like V100, A10G, or A5000."""

    GPU_HIGH = 3
    """Model requires high-end GPUs like A100 or H100 to achieve good throughput."""


@DeveloperAPI
@dataclass_json()
@dataclass
class ParameterMetadata:
    """Contains descriptive information that pertains to a Ludwig configuration parameter."""

    short_description: str = ""
    """Quick description generally for UI display."""

    long_description: str = ""
    """In depth description generally for documentation purposes."""

    ui_display_name: Union[str, None] = ""
    """How this parameter can be displayed in a human-readable form."""

    default_value_reasoning: Union[str, None] = None
    """The reasoning behind the default value for this parameter."""

    example_value: Union[List[Any], None] = None
    """Examples of other values that can be used for this parameter."""

    related_parameters: Union[List[str], None] = None
    """List of related parameters that this parameter interacts with or depends on."""

    other_information: Union[str, None] = None
    """Other information that is relevant for this parameter."""

    description_implications: Union[str, None] = None
    """The intuition for how model performance would change if this parameter is changed."""

    suggested_values: Any = None
    """What values would a machine learning expert suggest users try to help improve their model?

    Should cover 95% (2-sigma) worth of use-cases.
    """

    suggested_values_reasoning: Union[str, None] = None
    """The reasoning behind the suggested values, as well as model performance indicators or other intuition that
    could help inform a user to make an educated decision about what values to experiment with for this
    parameter."""

    commonly_used: bool = False
    """True if this parameter could be frequently used, would have a high impact, and/or would be interesting for a
    machine learning practitioner."""

    expected_impact: ExpectedImpact = ExpectedImpact.UNKNOWN
    """The expected impact of determining a "good" value for this parameter."""

    literature_references: Union[List[str], None] = None
    """List of links, papers, and blog posts to learn more."""

    internal_only: bool = False
    """True if this parameter is used strictly internally and should not be exposed to users."""

    compute_tier: ComputeTier = ComputeTier.CPU
    """The compute tier defines the type of compute resources that a model typically needs to get good
    throughput."""

    ui_component_type: Optional[str] = None
    """Override for HTML component type that should be used to render this field in UIs."""

    @memoized_method(maxsize=1)
    def to_json_dict(self) -> Dict[str, Any]:
        return json.loads(self.to_json())


@DeveloperAPI
def convert_metadata_to_json(pm: ParameterMetadata) -> Dict[str, Any]:
    """Converts a ParameterMetadata dict to a normal JSON dict.

    NOTE: Without the json.loads call, to_json() returns
    a string repr that is improperly parsed.
    """
    if not pm:
        return ParameterMetadata().to_json_dict()
    return pm.to_json_dict()


# This is a quick way to flag schema parameters as internal only via the `parameter_metadata` argument
INTERNAL_ONLY = ParameterMetadata(internal_only=True)



from ludwig.config_sampling.explore_schema import (
    combine_configs,
    combine_configs_for_comparator_combiner,
    combine_configs_for_sequence_combiner,
)

# A generic tabular to text config used to generate synthetic data and train a model on it.
TABULAR_TO_TEXT = """
input_features:
  - name: category_1
    type: category
  - name: number_1
    type: number
  - name: binary_1
    type: binary
output_features:
  - name: text_output_1
    type: text
"""

# A generic tabular config used to generate synthetic data and train a model on it.
TABULAR = """
input_features:
  - name: category_1
    type: category
  - name: number_1
    type: number
  - name: binary_1
    type: binary
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a single text input feature used to generate synthetic data and train a model on it.
TEXT_INPUT = """
input_features:
  - name: text_1
    type: text
    preprocessing:
      max_sequence_length: 8
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a single number input feature used to generate synthetic data and train a model on it.
NUMBER_INPUT = """
input_features:
  - name: number_1
    type: number
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a single category input feature used to generate synthetic data and train a model on it.
CATEGORY_INPUT = """
input_features:
  - name: category_1
    type: category
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a single binary input feature used to generate synthetic data and train a model on it.
BINARY_INPUT = """
input_features:
  - name: binary_1
    type: binary
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a text output feature used to generate synthetic data and train a model on it.
TEXT_OUTPUT = """
input_features:
  - name: text_1
    type: text
    preprocessing:
      max_sequence_length: 8
output_features:
  - name: text_output_1
    type: text
"""

# A generic config with a number output feature used to generate synthetic data and train a model on it.
NUMBER_OUTPUT = """
input_features:
  - name: number_1
    type: number
output_features:
  - name: number_output_1
    type: number
"""

# A generic config with a category output feature used to generate synthetic data and train a model on it.
CATEGORY_OUTPUT = """
input_features:
  - name: category_1
    type: category
output_features:
  - name: category_output_1
    type: category
"""

# A generic config with a binary output feature used to generate synthetic data and train a model on it.
BINARY_OUTPUT = """
input_features:
  - name: binary_1
    type: binary
output_features:
  - name: binary_output_1
    type: binary
"""

# Dictionary that maps from feature type to base config used to test the encoder and preprocessing sections.
FEATURE_TYPE_TO_CONFIG_FOR_ENCODER_PREPROCESSING = {
    "number": NUMBER_INPUT,
    "category": CATEGORY_INPUT,
    "binary": BINARY_INPUT,
    "text": TEXT_INPUT,
}

# Dictionary that maps from feature type to base config used to test the decoder and loss sections.
FEATURE_TYPE_TO_CONFIG_FOR_DECODER_LOSS = {
    "number": NUMBER_OUTPUT,
    "category": CATEGORY_OUTPUT,
    "binary": BINARY_OUTPUT,
    "text": TEXT_OUTPUT,
}

# Dictionary that maps from config section to base config used to test that section.
ECD_CONFIG_SECTION_TO_CONFIG = {
    "trainer": TABULAR,
    "comparator": TABULAR,
    "concat": TABULAR,
    "project_aggregate": TABULAR,
    "sequence": TEXT_INPUT,
    "sequence_concat": TEXT_INPUT,
    "tabnet": TABULAR,
    "tabtransformer": TABULAR,
    "transformer": TABULAR,
}

# Dictionary that maps from the combiner type to base config used to test that combiner.
COMBINER_TYPE_TO_COMBINE_FN_MAP = {
    "comparator": combine_configs_for_comparator_combiner,
    "concat": combine_configs,
    "project_aggregate": combine_configs,
    "sequence": combine_configs_for_sequence_combiner,
    "sequence_concat": combine_configs_for_sequence_combiner,
    "tabnet": combine_configs,
    "tabtransformer": combine_configs,
    "transformer": combine_configs,
}



import logging
from typing import Dict, Tuple

import torch
import torch.nn.functional as F
from bitsandbytes.nn.modules import Embedding
from transformers import (
    AutoModelForCausalLM,
    CodeLlamaTokenizer,
    CodeLlamaTokenizerFast,
    GPT2Tokenizer,
    GPT2TokenizerFast,
    LlamaTokenizer,
    LlamaTokenizerFast,
    PreTrainedTokenizer,
)

from ludwig.constants import IGNORE_INDEX_TOKEN_ID, LOGITS, PREDICTIONS, PROBABILITIES
from ludwig.schema.trainer import LLMTrainerConfig
from ludwig.utils.model_utils import find_embedding_layer_with_path

logger = logging.getLogger(__name__)


def set_pad_token(tokenizer: PreTrainedTokenizer):
    """Sets the pad token for the tokenizer if it is not already set.

    Args:
        tokenizer (PreTrainedTokenizer): The tokenizer.

    Example:
        >>> from transformers import GPT2Tokenizer, GPT2TokenizerFast, LlamaTokenizer, LlamaTokenizerFast # noqa
        >>> tokenizer = GPT2Tokenizer.from_pretrained('gpt2')
        >>> set_pad_token(tokenizer)
    """
    # Tokenizers might have the pad token id attribute since they tend to use the same base class, but
    # it can be set to None so we check for this explicitly.
    if hasattr(tokenizer, "pad_token_id") and tokenizer.pad_token_id is not None:
        return

    # HACK(Arnav): gpt, gpt2 and llama tokenizers had no pad tokens.
    # These recommend using eos tokens instead
    # https://github.com/huggingface/transformers/issues/2648#issuecomment-616177044
    # https://github.com/huggingface/transformers/issues/2630#issuecomment-1290809338
    if any(
        isinstance(tokenizer, t)
        for t in [
            GPT2Tokenizer,
            GPT2TokenizerFast,
            LlamaTokenizer,
            LlamaTokenizerFast,
            CodeLlamaTokenizer,
            CodeLlamaTokenizerFast,
        ]
    ):
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.pad_token_id = tokenizer.eos_token_id


def has_padding_token(input_tensor: torch.Tensor, tokenizer: PreTrainedTokenizer):
    """Checks if the input tensor contains any padding tokens.

    Args:
        input_tensor (torch.Tensor): The input tensor.
        tokenizer (PreTrainedTokenizer): The tokenizer used to encode the input.

    Returns:
        bool: True if the input tensor contains any padding tokens, False otherwise.

    Example:
        >>> import torch
        >>> from transformers import PreTrainedTokenizer
        >>> tokenizer = PreTrainedTokenizer.from_pretrained('bert-base-uncased')
        >>> input_sentence = "This is an example sentence."
        >>> input_ids = tokenizer.encode(input_sentence, add_special_tokens=True)
        >>> padded_input_ids = torch.nn.functional.pad(input_ids, (0, 10 - len(input_ids)))
        >>> has_padding = has_padding_token(padded_input_ids, tokenizer)
        >>> has_padding
        True
    """
    if input_tensor.dim() == 1:
        return torch.any(input_tensor == tokenizer.pad_token_id).item()
    elif input_tensor.dim() == 2:
        return torch.any(input_tensor == tokenizer.pad_token_id, dim=-1).item()
    else:
        raise ValueError("Input tensor must be 1D or 2D")


def remove_left_padding(input_ids_sample: torch.Tensor, tokenizer: PreTrainedTokenizer):
    """Removes left padding and other tokens until the first BOS token from the input_ids tensor.

    Args:
        input_ids_sample (torch.Tensor): The input tensor with padding and other tokens.
        tokenizer (PreTrainedTokenizer): The tokenizer used to encode the input.

    Returns:
        torch.Tensor: The input tensor without left padding and other tokens until the first BOS token.

    Example:
        >>> import torch
        >>> from transformers import PreTrainedTokenizer
        >>> tokenizer = PreTrainedTokenizer.from_pretrained('bert-base-uncased')
        >>> input_sentence = "This is an example sentence."
        >>> input_ids = tokenizer.encode(input_sentence, add_special_tokens=True)
        >>> padded_input_ids = torch.nn.functional.pad(input_ids, (10 - len(input_ids), 0))
        >>> input_ids_no_padding = remove_left_padding(padded_input_ids, tokenizer)
        >>> input_ids_no_padding
        tensor([[1, 2, 3]])
    """
    # Remove all PAD tokens
    pad_idxs = torch.where(input_ids_sample == tokenizer.pad_token_id)[0]  # all PAD token locations
    input_ids_no_padding = input_ids_sample
    if len(pad_idxs) != 0:
        pad_idx = pad_idxs[-1]  # get last PAD token location
        input_ids_no_padding = input_ids_sample[pad_idx + 1 :]

    # Start from the first BOS token
    bos_idxs = torch.where(input_ids_no_padding == tokenizer.bos_token_id)[0]  # all BOS token locations
    if len(bos_idxs) != 0:
        bos_idx = bos_idxs[0]  # get first BOS token location
    else:
        bos_idx = 0

    input_ids_no_bos = input_ids_no_padding[bos_idx:].unsqueeze(0)
    return input_ids_no_bos


def add_left_padding(input_ids, max_length, pad_value=0):
    """Adds left padding to the input_ids tensor.

    Args:
        input_ids (torch.Tensor): The input tensor.
        max_length (int): The maximum length of the tensor after padding.
        pad_value (int, optional): The value used for padding. Defaults to 0.

    Returns:
        torch.Tensor: The input_ids tensor with left padding.

    Example:
        >>> input_ids = torch.tensor([1, 2, 3])
        >>> max_length = 5
        >>> padded_tensor = add_left_padding(input_ids, max_length)
        >>> padded_tensor
        tensor([0, 0, 1, 2, 3])
    """
    padding = torch.tensor([pad_value] * (max_length - input_ids.shape[0]), dtype=torch.int64, device=input_ids.device)
    return torch.cat((padding, input_ids), dim=-1)


def create_attention_mask(input_ids: torch.Tensor, tokenizer: PreTrainedTokenizer):
    """Creates an attention mask for the input_ids tensor. This also sets the last padding token ID to 1 if it
    exists.

    Args:
        input_ids (torch.Tensor): The input tensor.
        tokenizer (PreTrainedTokenizer): The tokenizer used to encode the input.

    Returns:
        torch.Tensor: The attention mask tensor.

    Example:
        >>> import torch # noqa
        >>> from transformers import PreTrainedTokenizer
        >>> tokenizer = PreTrainedTokenizer.from_pretrained('bert-base-uncased')
        >>> input_sentence = "This is an example sentence."
        >>> input_ids = tokenizer.encode(input_sentence, add_special_tokens=True)
        >>> attention_mask = create_attention_mask(input_ids, tokenizer)
        >>> attention_mask
        tensor([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    """
    attention_mask = input_ids != tokenizer.pad_token_id
    # Last token may not be padding if we've already hit the max sequence length
    if not attention_mask[-1]:
        # last token is padding, always attended to even if it is padding
        attention_mask[-1] = 1
    attention_mask = attention_mask.to(torch.int64)
    return attention_mask


def find_last_matching_index(tensor_a: torch.Tensor, tensor_b: torch.Tensor):
    """Returns the last index of `tensor_a` that matches `tensor_b`. Specifically, this checks whether the tensor_b
    is in the last tensor_b.shape[0] elements of tensor_a.

    Args:
        tensor_a (torch.Tensor): The first tensor.
        tensor_b (torch.Tensor): The second tensor.

    Returns:
        int: The last index of `tensor_a` that matches `tensor_b`. Returns -1 if there is no matching index.

    Example:
        >>> import torch
        >>> tensor_a = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8])
        >>> tensor_b = torch.tensor([6, 7, 8])
        >>> last_matching_index = find_last_matching_index(tensor_a, tensor_b)
        >>> last_matching_index
        5
    """
    last_index = -1

    tensor_a_length = tensor_a.shape[0]
    tensor_b_length = tensor_b.shape[0]

    # Get the last tensor_b_length elements of tensor_a.
    tensor_a_truncated = tensor_a[-tensor_b_length:]

    # Find the last matching index.
    for i in range(tensor_b_length):
        if torch.equal(tensor_a_truncated[i:], tensor_b[: tensor_b_length - i]):
            last_index = tensor_a_length - tensor_b_length + i
            break

    return last_index


def pad_target_tensor_for_fine_tuning(
    targets: Dict[str, torch.Tensor],
    predictions: Dict[str, torch.Tensor],
    model_inputs: torch.Tensor,
    of_name: str,
) -> Dict[str, torch.Tensor]:
    """Pad and adjust target tensors for fine-tuning LLMS models.

    This function is used to pad and adjust the target tensors with IGNORE_INDEX_TOKEN_ID based on the model inputs and
    predictions during the fine-tuning process of Language Models. Here's what this function does:
        1. If none of the tokens from the target were in the model inputs, we create a tensor of the length of model
            inputs with value IGNORE_INDEX_TOKEN_IDs. This ignores this row from affecting loss.
        2. If the target tokens were entirely inside the model inputs, we want to pad all the tokens in model_inputs
            coming from the input with IGNORE_INDEX_TOKEN_IDs and leave the target tokens as is. This ensures that all
            of the target tokens are used during loss computation.
        3. In the scenario that only some part of the target tokens were in the model inputs, we want to pad the model
            inputs until that point and only leave the partial tokens of the target as is. This ensures that we will
            only compute loss on the target tokens that were in the model inputs.

    Args:
        targets (Dict[str, torch.Tensor]): A dictionary containing the target tensors.
        predictions (Dict[str, torch.Tensor]): A dictionary containing the predicted tensors.
        model_inputs (torch.Tensor): The input tensor passed into the model's forward pass.
        of_name (str): The name of the target tensor to be padded and adjusted.

    Returns:
        Dict[str, torch.Tensor]: A dictionary containing the updated target
        dictionaries.
    """
    target_length = targets.get(of_name).size()[1]
    prediction_length = predictions[of_name].get(PREDICTIONS).size()[1]

    if target_length == prediction_length:
        return targets

    updated_targets = []
    for idx, target in enumerate(targets[of_name]):
        # Remove any leading IGNORE_INDEX_TOKEN_IDs in the target that were temporarily added for alignment
        end_index = (target != IGNORE_INDEX_TOKEN_ID).nonzero()[0]
        target = target[end_index:]
        target_device = target.device

        # See if any part of the target was in the tensor passed into the model's forward pass
        last_matching_index = find_last_matching_index(model_inputs[idx], target)

        # If the last matching index is -1, it means that the input tensor passed into the model was truncated
        # and did not contain the target tensor. In this case, we need to truncate the target tensors as well
        # and just set it to a tensor of IGNORE_INDEX_TOKEN_ID so that we don't compute loss on this target tensor.
        if last_matching_index == -1:
            updated_targets.append(torch.full((prediction_length,), IGNORE_INDEX_TOKEN_ID).to(device=target_device))

        # If the last matching index is not -1, it means that the input tensor passed into the model was not
        # truncated and contained either a part of the target tensor or the entire target tensor. In this case,
        # we need to set the target tensor to the part of the target tensor that was passed into the model while
        # also padding it to the correct length with IGNORE_INDEX_TOKEN_ID.
        else:
            padding = torch.full((last_matching_index,), IGNORE_INDEX_TOKEN_ID).to(device=target_device)
            updated_targets.append(torch.cat((padding, target), dim=-1)[:prediction_length])

    targets[of_name] = torch.stack(updated_targets).to(device=targets.get(of_name).device, dtype=torch.int64)

    return targets


def generate_merged_ids(
    input_ids: torch.tensor, target_ids: torch.tensor, tokenizer: PreTrainedTokenizer, max_sequence_length: int = None
):
    """Generate merged input and target IDs tensor.

    This function merges the input_ids and target_ids together to create a unified tensor
    to pass into the model. It also returns attention masks for the merged tensors.

    Args:
        input_ids (torch.Tensor): The input IDs tensor.
        target_ids (torch.Tensor or None): The target IDs tensor or None.
        max_sequence_length (int or None): The maximum sequence length to pad or truncate to.
        tokenizer (PreTrainedTokenizer): The tokenizer used to encode the input_ids and target_ids.

    Returns:
        torch.Tensor: The merged input and target IDs tensor.
        torch.Tensor: The attention masks for the merged tensor.
    """
    merged_input_and_targets = []
    lengths = []

    pad_tensor = torch.tensor([tokenizer.pad_token_id]).to(target_ids[0].device)

    # Merge input_ids and target_ids by concatenating them together.
    # We remove the left padding from both input_ids and target_ids before concatenating them.
    for input_id_sample, target_id_sample in zip(input_ids, target_ids):
        input_id_sample_no_padding = remove_left_padding(input_id_sample, tokenizer)[0]
        target_id_sample_no_padding = remove_left_padding(target_id_sample, tokenizer)[0]
        target_id_sample_no_padding = torch.cat((target_id_sample_no_padding, pad_tensor), dim=-1)

        merged_sample_ids = torch.cat((input_id_sample_no_padding, target_id_sample_no_padding), dim=-1)
        # If the merged tensor is longer than the maximum sequence length, we truncate it.
        if max_sequence_length and merged_sample_ids.shape[0] > max_sequence_length:
            merged_sample_ids = merged_sample_ids[:max_sequence_length]

        merged_input_and_targets.append(merged_sample_ids)
        lengths.append(merged_sample_ids.shape[0])

    # Since we remove the left padding from the target_ids, the merged input_ids and target_ids
    # may not have the same lengths. We need to align them to the same length by adding left padding
    # and generate an attention mask for just the part of the input that is not padding.
    max_length = max(lengths)
    attention_masks = []
    for i, merged_sample_ids in enumerate(merged_input_and_targets):
        merged_input_and_targets[i] = add_left_padding(merged_sample_ids, max_length)
        attention_masks.append(create_attention_mask(merged_input_and_targets[i], tokenizer))

    return torch.stack(merged_input_and_targets), torch.stack(attention_masks)


def _get_decoded_targets_and_predictions(
    targets: Dict[str, torch.Tensor],
    predictions: Dict[str, Dict[str, torch.Tensor]],
    tokenizer: PreTrainedTokenizer,
    of_name: str,
):
    """Returns the decoded targets and predictions, accounting for IGNORE_INDEX_TOKEN_ID."""
    sanitized_targets = torch.where(targets[of_name] != IGNORE_INDEX_TOKEN_ID, targets[of_name], tokenizer.pad_token_id)
    sanitized_predictions = torch.where(
        predictions[of_name][PREDICTIONS] != IGNORE_INDEX_TOKEN_ID,
        predictions[of_name][PREDICTIONS],
        tokenizer.pad_token_id,
    )
    decoded_targets = tokenizer.batch_decode(sanitized_targets, skip_special_tokens=True)
    decoded_predictions = tokenizer.batch_decode(sanitized_predictions, skip_special_tokens=True)
    return decoded_targets, decoded_predictions


def realign_target_and_prediction_tensors_for_inference(
    targets: Dict[str, torch.Tensor],
    predictions: Dict[str, Dict[str, torch.Tensor]],
    of_name: str,
    tokenizer: PreTrainedTokenizer,
    pad_value: int = None,
) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
    """Realigns the target tensor with the predictions.

    This is necessary for text metrics that require the target and prediction to be of the same length.

    Args:
        targets: The target tensor.
        predictions: The prediction tensor.
        of_name: The output feature's name.
        tokenizer: The HF tokenizer.
        pad_direction: The direction to pad the tensors. Can be 'left' or 'right'.
            Defaults to 'right'.

    Returns:
        Tuple of realigned (targets, decoded_targets, predictions, decoded_predictions).
        - targets is a map of feature name -> tensor of token ids.
        - predictions is a map from output feature name -> map of tensors with the following items:
            - "predictions": tensor of token ids.
            - "probabilities": tensor of probabilities.
            - "logits": tensor of logits.
    """
    target_length = targets.get(of_name).size()[1]
    prediction_length = predictions[of_name].get(PREDICTIONS).size()[1]

    if target_length == prediction_length:
        return targets, predictions

    if not pad_value:
        pad_value = tokenizer.pad_token_id if tokenizer.pad_token_id is not None else tokenizer.eos_token_id

    # Align target and prediction tensors for text to text metric computation
    if target_length > prediction_length:
        # Pad the predictions.
        zeros_to_add = target_length - prediction_length

        predictions[of_name][PREDICTIONS] = F.pad(
            predictions[of_name][PREDICTIONS], (0, zeros_to_add), value=pad_value
        ).to(torch.int64)

        predictions[of_name][PROBABILITIES] = F.pad(predictions[of_name][PROBABILITIES], (0, 0, 0, zeros_to_add)).to(
            torch.float32
        )

        predictions[of_name][LOGITS] = F.pad(predictions[of_name][LOGITS], (0, 0, 0, zeros_to_add)).to(torch.float32)
    else:
        zeros_to_add = prediction_length - target_length

        targets[of_name] = F.pad(targets[of_name], (0, zeros_to_add), value=pad_value).to(torch.int64)

    return targets, predictions


def update_embedding_layer(model: AutoModelForCausalLM, config_obj: LLMTrainerConfig) -> AutoModelForCausalLM:
    """Updates the embedding layer of the model to use the 8-bit embedding layer from bitsandbytes.nn.modules.

    This is necessary when using 8-bit optimizers from bitsandbytes.
    See: https://github.com/TimDettmers/bitsandbytes#tldr
    """
    # If we're using an 8-bit optimizer, we need to replace the embedding layer with a custom embedding layer from
    # bnb.nn.modules.Embedding.
    if hasattr(config_obj, "optimizer") and config_obj.optimizer.is_8bit:
        embedding_layer, module_path = find_embedding_layer_with_path(model)
        if embedding_layer is None:
            raise ValueError(
                "Could not find an embedding layer in the model. This is required when using 8-bit optimizers"
                "  since a custom 8-bit embedding layer is used in place of the original embedding layer."
            )

        # Initialize the BNB embedding layer with the same parameters and weights as the original embedding layer.
        bnb_embedding = Embedding(
            num_embeddings=embedding_layer.num_embeddings,
            embedding_dim=embedding_layer.embedding_dim,
            padding_idx=embedding_layer.padding_idx,
            max_norm=embedding_layer.max_norm,
            norm_type=embedding_layer.norm_type,
            scale_grad_by_freq=embedding_layer.scale_grad_by_freq,
            sparse=embedding_layer.sparse,
            _weight=embedding_layer.weight,
            device=model.device,
        )

        # Update the model's original embedding layer to use the BNB embedding layer using the module_path
        # returned by find_embedding_layer_with_path.
        module_path = module_path.split(".")
        module = model
        for module_name in module_path[:-1]:
            module = getattr(module, module_name)
        setattr(module, module_path[-1], bnb_embedding)

        # Set the get input embeddings lambda function to return the BNB embedding layer
        model.get_input_embeddings = lambda: bnb_embedding

        logger.info("Updated the pretrained embedding layer to use the embedding layer from bitsandbytes.")

    return model



import logging
from typing import Type

from ludwig.api_annotations import DeveloperAPI
from ludwig.vector_index.base import VectorIndex

logger = logging.getLogger(__name__)


FAISS = "faiss"

ALL_INDICES = [FAISS]


def get_faiss_index_cls() -> Type[VectorIndex]:
    from ludwig.vector_index.faiss import FaissIndex

    return FaissIndex


# TODO(travis): add other indexing structures
vector_index_registry = {
    FAISS: get_faiss_index_cls,
}


@DeveloperAPI
def get_vector_index_cls(type: str) -> Type[VectorIndex]:
    return vector_index_registry[type]()



import numpy as np

from ludwig.api_annotations import PublicAPI
from ludwig.explain.explainer import Explainer
from ludwig.explain.explanation import ExplanationsResult
from ludwig.models.gbm import GBM


@PublicAPI(stability="experimental")
class GBMExplainer(Explainer):
    def explain(self) -> ExplanationsResult:
        """Explain the model's predictions. Uses the feature importances from the model.

        # Return

        :return: ExplanationsResult containing the explanations.
            `global_explanations`: (Explanation) Aggregate explanation for the entire input data.

            `row_explanations`: (List[Explanation]) A list of explanations, one for each row in the input data. Each
            explanation contains the feature attributions for each label in the target feature's vocab.

            `expected_values`: (List[float]) of length [output feature cardinality] Expected value for each label in
            the target feature's vocab.
        """
        base_model: GBM = self.model.model
        gbm = base_model.lgbm_model
        if gbm is None:
            raise ValueError("Model has not been trained yet.")

        # Get global feature importance from the model, use it for each row in the batch.
        # TODO(travis): support local feature importance
        raw_feat_imp = gbm.booster_.feature_importance(importance_type="gain")

        # For vector input features, the feature importance is given per element of the vector.
        # As such, to obtain the total importance for the feature, we need to sum over all the importance
        # values for every element of the vector.
        feat_imp = np.empty(len(base_model.input_features))
        raw_idx = 0
        for i, input_feature in enumerate(base_model.input_features.values()):
            # Length of the feature vector is the output shape of the encoder
            feature_length = input_feature.output_shape[0]
            raw_idx_end = raw_idx + feature_length

            # Reduce the importance values for every element in the vector down to the sum and
            # insert it as the feature level importance value
            feat_imp[i] = raw_feat_imp[raw_idx:raw_idx_end].sum()
            raw_idx = raw_idx_end

        # Logical check that at the end we reduced every element of the raw feature importance
        # into the feature level importance
        assert raw_idx == len(raw_feat_imp)

        # Scale the feature importance to sum to 1.
        feat_imp = feat_imp / feat_imp.sum() if feat_imp.sum() > 0 else feat_imp

        expected_values = []
        for _ in range(self.vocab_size):
            self.global_explanation.add(base_model.input_features.keys(), feat_imp)

            for explanation in self.row_explanations:
                # Add the feature attributions to the explanation object for this row.
                explanation.add(base_model.input_features.keys(), feat_imp)

            # TODO:
            expected_values.append(0.0)

        return ExplanationsResult(self.global_explanation, self.row_explanations, expected_values)



from typing import Any, Dict, Optional, Type

import deepspeed

from ludwig.backend.base import DataParallelBackend
from ludwig.constants import FALLBACK_BATCH_SIZE
from ludwig.distributed import init_dist_strategy
from ludwig.utils.batch_size_tuner import BatchSizeEvaluator


class DeepSpeedBackend(DataParallelBackend):
    BACKEND_TYPE = "deepspeed"

    def __init__(
        self,
        zero_optimization: Optional[Dict[str, Any]] = None,
        fp16: Optional[Dict[str, Any]] = None,
        bf16: Optional[Dict[str, Any]] = None,
        compression_training: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.zero_optimization = zero_optimization
        self.fp16 = fp16
        self.bf16 = bf16
        self.compression_training = compression_training

    def initialize(self):
        # Unlike when we use the Ray backend, we need to initialize the `torch.distributed` context so we can
        # broadcast, allgather, etc. before preparing the model within the trainer.
        deepspeed.init_distributed()
        self._distributed = init_dist_strategy(
            self.BACKEND_TYPE,
            zero_optimization=self.zero_optimization,
            fp16=self.fp16,
            bf16=self.bf16,
            compression_training=self.compression_training,
        )

    def supports_batch_size_tuning(self) -> bool:
        # TODO(travis): need to fix checkpoint saving/loading for DeepSpeed to enable tuning
        return False

    def tune_batch_size(self, evaluator_cls: Type[BatchSizeEvaluator], dataset_len: int) -> int:
        return FALLBACK_BATCH_SIZE



"""Checks that are not easily covered by marshmallow JSON schema validation like parameter interdependencies."""

from abc import ABC, abstractmethod
from re import findall
from typing import Callable, TYPE_CHECKING

from transformers import AutoConfig

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUDIO,
    BINARY,
    CATEGORY,
    IMAGE,
    IN_MEMORY,
    MIN_QUANTIZATION_BITS_FOR_MERGE_AND_UNLOAD,
    MODEL_ECD,
    MODEL_GBM,
    MODEL_LLM,
    NUMBER,
    SEQUENCE,
    SET,
    TEXT,
    TIMESERIES,
    VECTOR,
)
from ludwig.error import ConfigValidationError
from ludwig.utils.metric_utils import get_feature_to_metric_names_map_from_feature_collection
from ludwig.utils.misc_utils import merge_dict

if TYPE_CHECKING:
    from ludwig.schema.model_config import ModelConfig

# Set of all sequence feature types.
SEQUENCE_OUTPUT_FEATURE_TYPES = {SEQUENCE, TEXT, SET, VECTOR}


class ConfigCheckRegistry:
    """A registry of configuration checks."""

    def __init__(self):
        self._registry = []

    def register(self, check_fn):
        self._registry.append(check_fn)

    def check_config(self, config: "ModelConfig") -> None:  # noqa: F821
        for check_fn in self._registry:
            check_fn(config)


_CONFIG_CHECK_REGISTRY = ConfigCheckRegistry()


def get_config_check_registry():
    """Returns the config check registry."""
    return _CONFIG_CHECK_REGISTRY


@DeveloperAPI
def register_config_check(fn) -> Callable:
    """Registers a config check function."""
    _CONFIG_CHECK_REGISTRY.register(fn)


class ConfigCheck(ABC):
    """Checks instances of comprehensive (all parameters and defaults filled in) schema-validated config."""

    @staticmethod
    @abstractmethod
    def check(config: "ModelConfig") -> None:  # noqa: F821
        """Checks config for validity."""
        raise NotImplementedError


@register_config_check
def check_feature_names_unique(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that all feature names are unique."""
    input_features = config.input_features
    input_feature_names = {input_feature.name for input_feature in input_features}

    output_features = config.output_features
    output_feature_names = {output_feature.name for output_feature in output_features}

    if len(input_feature_names) + len(output_feature_names) != len(input_features) + len(output_features):
        raise ConfigValidationError("Feature names must be unique.")


@register_config_check
def check_tied_features_valid(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that all tied features are valid."""
    input_features = config.input_features
    input_feature_names = {input_feature.name for input_feature in input_features}

    for input_feature in input_features:
        if input_feature.tied and input_feature.tied not in input_feature_names:
            raise ConfigValidationError(
                f"Feature {input_feature.name} is tied to feature {input_feature.tied}, but the "
                f"'{input_feature.tied}' feature does not exist."
            )


@register_config_check
def check_training_runway(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that checkpoints_per_epoch and steps_per_checkpoint aren't simultaneously defined."""
    if config.model_type == MODEL_ECD:
        if config.trainer.checkpoints_per_epoch != 0 and config.trainer.steps_per_checkpoint != 0:
            raise ConfigValidationError(
                "It is invalid to specify both trainer.checkpoints_per_epoch AND "
                "trainer.steps_per_checkpoint. Please specify one or the other, or specify neither to "
                "checkpoint/eval the model every epoch."
            )


@register_config_check
def check_gbm_horovod_incompatibility(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that GBM model type isn't being used with the horovod backend.

    TODO(Justin): This is fine for now because we don't validate on the backend, but can be removed in the future when
    backend is schema-fied (separate schemas for ECD and GBM).
    """
    if config.backend is None:
        return
    # TODO (jeffkinnison): Revert to object access when https://github.com/ludwig-ai/ludwig/pull/3127 lands
    if config.model_type == MODEL_GBM and config.backend.get("type") == "horovod":
        raise ConfigValidationError("Horovod backend does not support GBM models.")


@register_config_check
def check_gbm_output_type(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that the output features for GBMs are of supported types."""
    if config.model_type == MODEL_GBM:
        for output_feature in config.output_features:
            if output_feature.type not in {BINARY, CATEGORY, NUMBER}:
                raise ConfigValidationError(
                    "GBM Models currently only support Binary, Category, and Number output features."
                )


@register_config_check
def check_ray_backend_in_memory_preprocessing(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that in memory preprocessing is used with Ray backend."""
    if config.backend is None:
        return
    if not hasattr(config.trainer, "preprocessing") or not hasattr(config.trainer.preprocessing, IN_MEMORY):
        return

    if config.backend.type == "ray" and not config.trainer.preprocessing.in_memory:
        raise ConfigValidationError(
            "RayBackend does not support lazy loading of data files at train time. "
            "Set preprocessing config `in_memory: True`"
        )

    for input_feature in config.input_features:
        if input_feature.type == AUDIO or input_feature.type == IMAGE:
            if not input_feature.preprocessing.in_memory and config.backend.type != "ray":
                raise ConfigValidationError(
                    "RayBackend does not support lazy loading of data files at train time. "
                    f"Set preprocessing config `in_memory: True` for input feature {input_feature.name}"
                )


def check_sequence_concat_combiner_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that sequence concat combiner has at least one input feature that's sequential."""
    if config.model_type != MODEL_ECD:
        return
    if config.combiner != "sequence_concat":
        return
    has_sequence_input = False
    for input_feature in config.input_features:
        if input_feature.type in SEQUENCE_OUTPUT_FEATURE_TYPES:
            has_sequence_input = True
            break
    if not has_sequence_input:
        raise ConfigValidationError(
            "Sequence concat combiner should only be used for at least one sequential input feature."
        )


@register_config_check
def check_comparator_combiner_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that all of the feature names for entity_1 and entity_2 are valid features."""
    if config.model_type != MODEL_ECD:
        return
    if config.combiner.type != "comparator":
        return

    input_feature_names = [input_feature.name for input_feature in config.input_features]
    for feature_name in config.combiner.entity_1:
        if feature_name not in input_feature_names:
            raise ConfigValidationError(
                f"Feature {feature_name} in entity_1 for the comparator combiner is not a valid " "input feature name."
            )
    for feature_name in config.combiner.entity_2:
        if feature_name not in input_feature_names:
            raise ConfigValidationError(
                f"Feature {feature_name} in entity_2 for the comparator combiner is not a valid " "input feature name."
            )

    if sorted(config.combiner.entity_1 + config.combiner.entity_2) != sorted(input_feature_names):
        raise ConfigValidationError("Not all input features are present as entities in the comparator combiner.")


@register_config_check
def check_class_balance_preprocessing(config: "ModelConfig") -> None:  # noqa: F821
    """Class balancing is only available for datasets with a single output feature."""
    if config.preprocessing.oversample_minority or config.preprocessing.undersample_majority:
        if len(config.output_features) != 1:
            raise ConfigValidationError("Class balancing is only available for datasets with a single output feature.")
        if config.output_features[0].type != BINARY:
            raise ConfigValidationError("Class balancing is only supported for binary output features.")


@register_config_check
def check_sampling_exclusivity(config: "ModelConfig") -> None:  # noqa: F821
    """Oversample minority and undersample majority are mutually exclusive."""
    if config.preprocessing.oversample_minority and config.preprocessing.undersample_majority:
        raise ConfigValidationError(
            "Oversample minority and undersample majority are mutually exclusive. Specify only one method."
        )


@register_config_check
def check_validation_metric_exists(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that the specified validation metric exists."""
    validation_metric_name = config.trainer.validation_metric

    # Get all valid metrics.
    feature_to_metric_names_map = get_feature_to_metric_names_map_from_feature_collection(config.output_features)
    all_valid_metrics = set()
    for metric_names in feature_to_metric_names_map.values():
        all_valid_metrics.update(metric_names)

    if validation_metric_name not in all_valid_metrics:
        raise ConfigValidationError(
            f"User-specified trainer.validation_metric '{validation_metric_name}' is not valid. "
            f"Available metrics are: {all_valid_metrics}"
        )


@register_config_check
def check_splitter(config: "ModelConfig") -> None:  # noqa: F821
    """Checks the validity of the splitter configuration."""
    from ludwig.data.split import get_splitter

    splitter = get_splitter(**config.preprocessing.split.to_dict())
    splitter.validate(config)


@register_config_check
def check_hf_tokenizer_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that the HuggingFace tokenizer has a pretrained_model_name_or_path specified."""

    for input_feature in config.input_features:
        if input_feature.type == TEXT:
            if input_feature.preprocessing.tokenizer == "hf_tokenizer":
                if input_feature.preprocessing.pretrained_model_name_or_path is None:
                    raise ConfigValidationError(
                        "Pretrained model name or path must be specified for HuggingFace tokenizer."
                    )


@register_config_check
def check_hf_encoder_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that a HuggingFace encoder has a pretrained_model_name_or_path specified."""

    for input_feature in config.input_features:
        if input_feature.type == TEXT:
            if hasattr(input_feature.encoder, "use_pretrained"):
                if input_feature.preprocessing.pretrained_model_name_or_path is None:
                    raise ConfigValidationError(
                        "Pretrained model name or path must be specified for HuggingFace encoder."
                    )


@register_config_check
def check_stacked_transformer_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that the transformer encoder type correctly configures `num_heads` and `hidden_size`"""

    def is_divisible(hidden_size: int, num_heads: int) -> bool:
        """Checks that hidden_size is divisible by num_heads."""
        return hidden_size % num_heads == 0

    sequence_types = [SEQUENCE, TEXT, TIMESERIES]

    for input_feature in config.input_features:
        if_type = input_feature.type
        encoder = input_feature.encoder
        if (
            if_type in sequence_types
            and encoder.type == "transformer"
            and not is_divisible(encoder.hidden_size, encoder.num_heads)
        ):
            raise ConfigValidationError(
                f"Input feature {input_feature.name} transformer encoder requires encoder.hidden_size to be divisible "
                f"by encoder.num_heads. Found hidden_size {encoder.hidden_size} and num_heads {encoder.num_heads}."
            )


@register_config_check
def check_hyperopt_search_algorithm_dependencies_installed(config: "ModelConfig") -> None:  # noqa: F821
    """Check that the hyperopt search algorithm dependencies are installed."""
    if config.hyperopt is None:
        return

    try:
        config.hyperopt.search_alg.dependencies_installed()
    except ImportError as e:
        raise ConfigValidationError(e.msg)


@register_config_check
def check_hyperopt_scheduler_dependencies_installed(config: "ModelConfig") -> None:  # noqa: F821
    """Check that the hyperopt scheduler dependencies are installed."""
    if config.hyperopt is None:
        return

    try:
        config.hyperopt.executor.scheduler.dependencies_installed()
    except ImportError as e:
        raise ConfigValidationError(e.msg)


@register_config_check
def check_tagger_decoder_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that the tagger decoder has at least one sequence, text or timeseries input feature where the
    encoder's reduce_output will produce a 3D shaped output from the combiner."""
    # Check if there is a text or sequence output feature using a tagger decoder
    output_feature_with_tagger_decoder = False
    for output_feature in config.output_features:
        if output_feature.type in {TEXT, SEQUENCE} and output_feature.decoder.type == "tagger":
            output_feature_with_tagger_decoder = True

    if not output_feature_with_tagger_decoder:
        return

    # Check that there is at least one sequence, text or timeseries input feature that doesn't reduce the
    # output of the encoder.
    has_sequence_feature = False
    for input_feature in config.input_features:
        if input_feature.type in {SEQUENCE, TEXT, TIMESERIES}:
            has_sequence_feature = True
            if input_feature.encoder.reduce_output is None:
                return

    if not has_sequence_feature:
        raise ConfigValidationError("Tagger decoder requires at least one text, sequence or timeseries input feature.")
    else:
        raise ConfigValidationError(
            "Tagger decoder requires at least one of the text, sequence or timeseries input feature encoders to have "
            "`reduce_output` set to `None`."
        )


@register_config_check
def check_hyperopt_parameter_dicts(config: "ModelConfig") -> None:  # noqa: F821
    """Checks for hyperopt parameter dicts against their config objects."""
    if config.hyperopt is None:
        return

    from ludwig.schema.hyperopt.utils import get_parameter_cls, parameter_config_registry  # noqa: F401

    for parameter, space in config.hyperopt.parameters.items():
        # skip nested hyperopt parameters
        if parameter != ".":
            parameter_attribute_path = parameter.split(".")
            passed = False

            for root in [config, config.input_features, config.output_features]:
                current = root
                for p in parameter_attribute_path:
                    try:
                        current = current.__getattribute__(p)
                        if p == parameter_attribute_path[-1]:
                            passed = True
                    except AttributeError:
                        break
                if passed:
                    break

            if not passed:
                raise ConfigValidationError(
                    f"The supplied hyperopt parameter {parameter} is not a valid config field. Check the Ludwig "
                    "docs for the list of valid parameters."
                )

            try:
                space_cls = get_parameter_cls(space["space"])
                space_cls.from_dict(space)
            except KeyError:
                space_types = ", ".join(parameter_config_registry.keys())
                raise ConfigValidationError(
                    f"Invalid hyperopt parameter space requested for `hyperopt.parameters.{parameter}`. Valid spaces "
                    f"are {space_types}."
                )


@register_config_check
def check_concat_combiner_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that if the concat combiner receives a mixture of sequence and non-sequence features, that all
    sequence features are configured with reduce_output to be 2D tensors."""
    if config.model_type != MODEL_ECD:
        return
    if config.combiner.type != "concat":
        return

    has_unreduced_sequence_feature = False
    has_non_sequence_feature = False
    for input_feature in config.input_features:
        if (
            input_feature.type in {SEQUENCE, TEXT, TIMESERIES}
            and hasattr(input_feature.encoder, "reduce_output")
            and input_feature.encoder.reduce_output is None
        ):
            has_unreduced_sequence_feature = True
        else:
            has_non_sequence_feature = True

    if has_unreduced_sequence_feature and has_non_sequence_feature:
        raise ConfigValidationError(
            "The concat combiner cannot receive a mix of unreduced sequence features (3D) and non-sequence features "
            "(2D). Options: 1) Set reduce_output in sequence feature encoders to a value other than None to ensure 2D "
            "encoder outputs, 2) Choose a different combiner like `sequence_concat` which can handle a mix of 2D and "
            "3D encoder output shapes, or 3) Remove features to ensure that output shapes from all encoders are the "
            "same dimension (all 2D or all 3D)."
        )


@register_config_check
def check_hyperopt_nested_parameter_dicts(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that all nested parameters in a hyperopt config exist."""
    if config.hyperopt is None or "." not in config.hyperopt.parameters:
        return

    from ludwig.schema.hyperopt.utils import get_parameter_cls  # noqa: F401
    from ludwig.schema.model_types.base import ModelConfig

    space = config.hyperopt.parameters["."]

    # Build the config that would be produced by each parameter dict to validate subsections that may be in
    config_dict = config.to_dict()
    del config_dict["hyperopt"]
    for category in space["categories"]:
        for i, k in enumerate(category.keys()):
            try:
                config.__getattribute__(k)
            except AttributeError:
                raise ConfigValidationError(f"Invalid config block {k} in nested hyperopt parameter dict {i}: {space}.")

        category_dict = merge_dict(config_dict, category)
        try:
            ModelConfig.from_dict(category_dict)
        except ConfigValidationError as e:
            raise ConfigValidationError(f"Invalid config in hyperopt nested parameter config: {category}. {e.message}")

    try:
        space_cls = get_parameter_cls("choice")
        space_cls.from_dict(space)
    except KeyError:
        raise ConfigValidationError(
            f"Nested hyperparameter search spaces must be of type 'choice'. Requested space type: {space['space']}"
        )


@register_config_check
def check_llm_exactly_one_input_text_feature(config: "ModelConfig"):  # noqa: F821
    if config.model_type != MODEL_LLM:
        return

    if len(config.input_features) == 1 and config.input_features[0].type == TEXT:
        return
    else:
        raise ConfigValidationError("LLM requires exactly one text input feature.")


@register_config_check
def check_llm_finetuning_output_feature_config(config: "ModelConfig"):  # noqa: F821
    """Checks that the output feature config for LLM finetuning is valid."""
    if config.model_type != MODEL_LLM:
        return

    if config.trainer.type != "finetune":
        return

    if config.output_features[0].type != TEXT:
        raise ConfigValidationError(
            "LLM finetuning requires the output feature to be a text feature. If you are trying to use a different "
            "output feature type such as category or binary, please change the output feature type to text."
        )


@register_config_check
def check_llm_finetuning_trainer_config(config: "ModelConfig"):  # noqa: F821
    """Ensures that trainer type is finetune if adapter is not None."""
    if config.model_type != MODEL_LLM:
        return

    if (
        config.trainer.type == "none"
        and config.adapter is not None
        and config.adapter.pretrained_adapter_weights is not None
    ):
        # If performing zero-shot, we must specify pretrained adapter weights
        return

    if config.adapter is not None and config.trainer.type != "finetune":
        raise ConfigValidationError("LLM finetuning requires trainer type to be finetune.")


@register_config_check
def check_llm_finetuning_backend_config(config: "ModelConfig"):  # noqa: F821
    """Checks that the LLM finetuning using Ray is configured correctly.

    DDP strategy is not supported for LLM finetuning because it leads to OOMs since the model is large and DDP strategy
    requires a copy of the model on each GPU.
    """
    if config.model_type != MODEL_LLM:
        return

    # LLM finetuning is only supported by the finetune trainer type
    if (
        config.trainer.type != "finetune"
        and config.adapter is not None
        and config.adapter.pretrained_adapter_weights is not None
    ):
        return

    # Using local backend, so skip the checks below
    if not hasattr(config.backend, "type"):
        return

    backend = config.backend
    if not hasattr(backend.trainer, "strategy") or backend.trainer.strategy != "deepspeed":
        raise ConfigValidationError("LLM finetuning with Ray requires the DeepSpeed strategy.")

    # Deepspeed requires GPU
    if not backend.trainer.use_gpu or backend.trainer.resources_per_worker.GPU < 1:
        raise ConfigValidationError("LLM finetuning with DeepSpeed requires GPU.")


@register_config_check
def check_llm_finetuning_adalora_config(config: "ModelConfig"):
    """Checks that the adalora adapter is configured correctly.

    We check against PEFT's predefined target module list for ADALORA to see if this target_modules is present there. If
    not, AdaloraModel will run into issues downstream.
    """
    if config.model_type != MODEL_LLM:
        return

    if not config.adapter:
        return

    if config.adapter.type != "adalora":
        return

    from peft.utils import TRANSFORMERS_MODELS_TO_ADALORA_TARGET_MODULES_MAPPING

    model_config = _get_llm_model_config(config.base_model)
    if model_config.model_type not in TRANSFORMERS_MODELS_TO_ADALORA_TARGET_MODULES_MAPPING:
        raise ConfigValidationError(
            f"Adalora adapter is not supported for {model_config.model_type} model. "
            f"Supported model types are: {list(TRANSFORMERS_MODELS_TO_ADALORA_TARGET_MODULES_MAPPING.keys())}. "
            "If you know the target modules for your model, please specify them in the config through the "
            "`target_modules` key."
        )


@register_config_check
def check_llm_finetuning_adaption_prompt_parameters(config: "ModelConfig"):
    """Checks that the adaption_prompt adapter is configured correctly.

    Adaption prompt is only supported for Llama models.
    """
    if config.model_type != MODEL_LLM:
        return

    if not config.adapter:
        return

    if config.adapter.type != "adaption_prompt":
        return

    from peft.tuners.adaption_prompt import TRANSFORMERS_MODEL_CONFIG

    # Adaption Config is currently only supported for Llama model types
    model_config = _get_llm_model_config(config.base_model)
    if model_config.model_type not in TRANSFORMERS_MODEL_CONFIG:
        raise ConfigValidationError(
            f"Adaption prompt adapter is not supported for {model_config.model_type} model. "
            f"Supported model types are: {list(TRANSFORMERS_MODEL_CONFIG.keys())}."
        )


def _get_llm_model_config(model_name: str) -> AutoConfig:
    """Returns the LLM model config."""
    return AutoConfig.from_pretrained(model_name)


# TODO(geoffrey, arnav): uncomment this when we have reconciled the config with the backend kwarg in api.py
# @register_config_check
def check_llm_quantization_backend_incompatibility(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that LLM model type with quantization uses the local backend."""
    if config.model_type != MODEL_LLM:
        return

    if config.quantization is None:
        return

    backend_type = None
    if config.backend:
        backend_type = config.backend.get("type", None)

    # If backend was explicitly set to Ray, then we need to raise an error
    if backend_type == "ray":
        raise ConfigValidationError(f"LLM with quantization requires the 'local' backend, found: '{backend_type}'")

    # If the backend is not explicitly set, then we need to check if a Ray process is running
    # If a Ray process is running, then we need to raise an error because the backend will be set to Ray
    if config.backend is None:
        try:
            # May not be installed, so we need to catch the ImportError
            import ray

            if ray.is_initialized():
                raise ConfigValidationError(
                    "LLM with quantization requires the 'local' backend, but backend will be set "
                    "to Ray since Ray is already running locally."
                )
        except ImportError:
            pass


@register_config_check
def check_qlora_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that all the necessary settings are in place for QLoRA."""
    if config.model_type != MODEL_LLM or config.trainer.type == "none":
        return

    if config.quantization and (not config.adapter or config.adapter.type != "lora"):
        raise ConfigValidationError("Fine-tuning and LLM with quantization requires using the 'lora' adapter")


@register_config_check
def check_qlora_merge_and_unload_compatibility(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that model.merge_and_unload() is supported by underlying model.save_pretrained() when merging QLoRA
    layers."""
    if config.model_type != MODEL_LLM or config.trainer.type == "none":
        return

    if not (
        config.adapter
        and config.adapter.type in ["lora", "adalora"]
        and config.adapter.postprocessor
        and config.adapter.postprocessor.merge_adapter_into_base_model
        and config.quantization
    ):
        return

    if config.quantization.bits < MIN_QUANTIZATION_BITS_FOR_MERGE_AND_UNLOAD:
        raise ConfigValidationError(
            f"""This operation will entail merging LoRA layers on a {config.quantization.bits}-bit \
quantized model.  Calling "save_pretrained()" on that model is currently unsupported.  If you want to merge the LoRA \
adapter weights into the base model, you need to use 8-bit quantization or do non-quantized based training by removing \
the quantization section from your Ludwig configuration."""
        )


@register_config_check
def check_prompt_requirements(config: "ModelConfig") -> None:  # noqa: F821
    """Checks that prompt's template and task properties are valid, according to the description on the schema."""
    if config.model_type != MODEL_LLM:
        return

    # TODO: `prompt` by default should be set to null, not a default dict:
    # # If no prompt is provided, no validation necessary:
    # if not config.prompt:
    #     return
    from ludwig.schema.llms.prompt import PromptConfig, RetrievalConfig

    if config.prompt == PromptConfig():
        return

    template = config.prompt.template
    task = config.prompt.task
    retrieval = config.prompt.retrieval

    # If template is NOT provided, then task is required for zero/few shot learning:
    if not template and not task:
        raise ConfigValidationError("A prompt task is required if no template is provided!")

    template_refs = set(findall(r"\{(.*?)\}", template)) if isinstance(template, str) else set()

    # If a template IS provided (i.e. we are not doing a built-in zero/few-shot learning), then...
    if template:
        # If task is also provided, the template must contain it:
        if task and "__task__" not in template_refs:
            raise ConfigValidationError(
                "When providing a task, you must make sure that the task keyword `{__task__} is "
                "present somewhere in the template string!"
            )

        # If retrieval is also provided, the template must reference it:
        # TODO: retrieval by default should be set to null, not a default dict:
        if retrieval and retrieval != RetrievalConfig() and "__context__" not in template_refs:
            raise ConfigValidationError(
                "When providing a retrieval config, you must make sure that the task keyword `{__context__}` is "
                "present somewhere in the template string!"
            )

        # Otherwise, the template should at least contain the sample keyword or some input column:
        # TODO: len(template_refs) is a hacky attempt to check that there are references to *something* in the
        # string. The proper validation is to check the references against the features in the user's dataset - but we
        # do not have access to the dataset in this code path right now.
        if not task:
            if len(template_refs) == 0 and "__sample__" not in template_refs:
                raise ConfigValidationError(
                    "A template must contain at least one reference to a column or the sample keyword {__sample__} for "
                    "a JSON-serialized representation of non-output feature columns."
                )



from typing import Callable, List, Optional, Tuple, Type

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.registry import Registry

parameter_config_registry = Registry()
scheduler_config_registry = Registry()
scheduler_dependencies_registry = Registry()
search_algorithm_config_registry = Registry()
search_algorithm_dependencies_registry = Registry()
search_algorithm_random_state_field_registry = Registry()


@DeveloperAPI
def get_parameter_cls(name: str) -> Type["BaseParameterConfig"]:  # noqa: F821
    """Get a registered hyperopt parameter config class by name.

    Args:
        name: the name of a parameter config class registered in `ludwig.schema.hyperopt.parameter`

    Returns:
        A parameter config class from `ludwig.schema.hyperopt.parameter`
    """
    return parameter_config_registry[name]


@DeveloperAPI
def get_scheduler_cls(name: str) -> Type["BaseSchedulerConfig"]:  # noqa: F821
    """Get a registered hyperopt scheduler config class by name.

    Args:
        name: the name of a scheduler config class registered in `ludwig.schema.hyperopt.scheduler`

    Returns:
        A scheduler config class from `ludwig.schema.hyperopt.scheduler`
    """
    return search_algorithm_config_registry[name]


@DeveloperAPI
def get_scheduler_dependencies(name: str) -> List[str]:
    """Get the list of dependencies for a registered hyperopt scheduler.

    Args:
        name: the name of a scheduler config class registered in `ludwig.schema.hyperopt.scheduler`

    Returns:
        The list of imports needed to use the scheduler
    """
    return scheduler_dependencies_registry[name]


@DeveloperAPI
def get_search_algorithm_cls(name: str) -> Type["BaseSearchAlgorithmConfig"]:  # noqa: F821
    """Get a registered hyperopt search algorithm config class by name.

    Args:
        name: the name of a search algorithm config class registered in `ludwig.schema.hyperopt.search_algorithm`

    Returns:
        A scheduler config class from `ludwig.schema.hyperopt.search_algorithm`
    """
    return search_algorithm_config_registry[name]


@DeveloperAPI
def get_search_algorithm_dependencies(name: str) -> List[str]:
    """Get the list of dependencies for a registered hyperopt search algorithm.

    Args:
        name: the name of a search algorithm config class registered in `ludwig.schema.hyperopt.search_algorithm`

    Returns:
        The list of imports needed to use the search algorithm
    """
    return search_algorithm_dependencies_registry[name]


@DeveloperAPI
def get_search_algorithm_random_state_field(name: str):
    """Get the field name of the random state for a registered hyperopt search algorithm.

    Args:
        name: the name of a search algorithm config class registered in `ludwig.schema.hyperopt.search_algorithm`

    Returns:
        The name of the random state field in the config
    """
    return search_algorithm_random_state_field_registry[name]


@DeveloperAPI
def register_parameter_config(name: str) -> Callable:
    """Register a parameter config class by name.

    Args:
        name: the name to register the parameter class under, does not need to correspond to the value of `space`

    Returns:
        Wrapper function to decorate a `BaseParameterConfig` subclass
    """

    def wrap(cls: Type["BaseParameterConfig"]) -> Type["BaseParameterConfig"]:  # noqa: F821
        """Add a parameter config class to the registry.

        Args:
            cls: a subclass of `BaseParameterConfig`

        Returns:
            `cls` unaltered
        """
        parameter_config_registry[name] = cls
        return cls

    return wrap


@DeveloperAPI
def register_scheduler_config(name: str, dependencies: Optional[List[Tuple[str]]] = None):
    """Register a scheduler config class by name.

    Args:
        name: the name to scheduler the parameter class under, does not need to correspond to the value of `type`
        dependencies: the list of scheduler dependency package name/install name pairs, e.g.
                      `("sklearn", "scikit-learn")`

    Returns:
        Wrapper function to decorate a `BaseSchedulerConfig` subclass
    """

    def wrap(scheduler_config: Type["BaseSchedulerConfig"]) -> Type["BaseSchedulerConfig"]:  # noqa: F821
        """Add a parameter config class to the registry.

        Args:
            cls: a subclass of `BaseParameterConfig`

        Returns:
            `cls` unaltered
        """
        scheduler_config_registry[name] = scheduler_config
        scheduler_dependencies_registry[name] = dependencies if dependencies is not None else []
        return scheduler_config

    return wrap


# TODO: create a search alg metadata class to register in place of individual metadata args
@DeveloperAPI
def register_search_algorithm_config(
    name: str, random_state_field: Optional[str] = None, dependencies: Optional[List[Tuple[str, str]]] = None
) -> Callable:
    """Register a search algorithm config class by name.

    Args:
        name: the name to register the search algorithm class under, does not need to correspond to the value of `type`
        random_state_field: the name of the random state in this search algorithm
        dependencies: the list of search algorithm dependency package name/install name pairs, e.g.
                      `("sklearn", "scikit-learn")`

    Returns:
        Wrapper function to decorate a `BaseSearchAlgorithmConfig` subclass
    """

    def wrap(cls: Type["BaseSearchAlgorithmConfig"]) -> Type["BaseSearchAlgorithmConfig"]:  # noqa: F821
        search_algorithm_config_registry[name] = cls
        search_algorithm_dependencies_registry[name] = dependencies if dependencies is not None else []
        search_algorithm_random_state_field_registry[name] = random_state_field
        return cls

    return wrap



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from __future__ import annotations

import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Callable, TYPE_CHECKING

import numpy as np
import pandas as pd
import psutil
import torch
from tqdm import tqdm

from ludwig.api_annotations import DeveloperAPI
from ludwig.backend.utils.storage import StorageManager
from ludwig.constants import MODEL_LLM
from ludwig.data.cache.manager import CacheManager
from ludwig.data.dataframe.base import DataFrameEngine
from ludwig.data.dataframe.pandas import PANDAS
from ludwig.data.dataset.base import DatasetManager
from ludwig.data.dataset.pandas import PandasDatasetManager
from ludwig.distributed import init_dist_strategy
from ludwig.distributed.base import DistributedStrategy
from ludwig.models.base import BaseModel
from ludwig.schema.trainer import BaseTrainerConfig
from ludwig.types import HyperoptConfigDict
from ludwig.utils.audio_utils import read_audio_from_path
from ludwig.utils.batch_size_tuner import BatchSizeEvaluator
from ludwig.utils.dataframe_utils import from_batches, to_batches
from ludwig.utils.fs_utils import get_bytes_obj_from_path
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.system_utils import Resources
from ludwig.utils.torch_utils import initialize_pytorch
from ludwig.utils.types import DataFrame, Series

if TYPE_CHECKING:
    from ludwig.trainers.base import BaseTrainer


@DeveloperAPI
class Backend(ABC):
    def __init__(
        self,
        dataset_manager: DatasetManager,
        cache_dir: str | None = None,
        credentials: dict[str, dict[str, Any]] | None = None,
    ):
        credentials = credentials or {}
        self._dataset_manager = dataset_manager
        self._storage_manager = StorageManager(**credentials)
        self._cache_manager = CacheManager(self._dataset_manager, cache_dir)

    @property
    def storage(self) -> StorageManager:
        return self._storage_manager

    @property
    def cache(self) -> CacheManager:
        return self._cache_manager

    @property
    def dataset_manager(self) -> DatasetManager:
        return self._dataset_manager

    @abstractmethod
    def initialize(self):
        raise NotImplementedError()

    @abstractmethod
    def initialize_pytorch(self, *args, **kwargs):
        raise NotImplementedError()

    @contextmanager
    @abstractmethod
    def create_trainer(self, **kwargs) -> BaseTrainer:
        raise NotImplementedError()

    @abstractmethod
    def sync_model(self, model):
        raise NotImplementedError()

    @abstractmethod
    def broadcast_return(self, fn):
        raise NotImplementedError()

    @abstractmethod
    def is_coordinator(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def df_engine(self) -> DataFrameEngine:
        raise NotImplementedError()

    @property
    @abstractmethod
    def supports_multiprocessing(self):
        raise NotImplementedError()

    @abstractmethod
    def read_binary_files(self, column: Series, map_fn: Callable | None = None) -> Series:
        raise NotImplementedError()

    @property
    @abstractmethod
    def num_nodes(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def num_training_workers(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def get_available_resources(self) -> Resources:
        raise NotImplementedError()

    @abstractmethod
    def max_concurrent_trials(self, hyperopt_config: HyperoptConfigDict) -> int | None:
        raise NotImplementedError()

    @abstractmethod
    def tune_batch_size(self, evaluator_cls: type[BatchSizeEvaluator], dataset_len: int) -> int:
        """Returns best batch size (measured in samples / s) on the given evaluator.

        The evaluator class will need to be instantiated on each worker in the backend cluster, then call
        `evaluator.select_best_batch_size(dataset_len)`.
        """
        raise NotImplementedError()

    @abstractmethod
    def batch_transform(self, df: DataFrame, batch_size: int, transform_fn: Callable, name: str = None) -> DataFrame:
        """Applies `transform_fn` to every `batch_size` length batch of `df` and returns the result."""
        raise NotImplementedError()

    def supports_batch_size_tuning(self) -> bool:
        return True


class LocalPreprocessingMixin:
    @property
    def df_engine(self):
        return PANDAS

    @property
    def supports_multiprocessing(self):
        return True

    @staticmethod
    def read_binary_files(column: pd.Series, map_fn: Callable | None = None, file_size: int | None = None) -> pd.Series:
        column = column.fillna(np.nan).replace([np.nan], [None])  # normalize NaNs to None

        sample_fname = column.head(1).values[0]
        with ThreadPoolExecutor() as executor:  # number of threads is inferred
            if isinstance(sample_fname, str):
                if map_fn is read_audio_from_path:  # bypass torchaudio issue that no longer takes in file-like objects
                    result = executor.map(lambda path: map_fn(path) if path is not None else path, column.values)
                else:
                    result = executor.map(
                        lambda path: get_bytes_obj_from_path(path) if path is not None else path, column.values
                    )
            else:
                # If the sample path is not a string, assume the paths has already been read in
                result = column.values

            if map_fn is not None and map_fn is not read_audio_from_path:
                result = executor.map(map_fn, result)

        return pd.Series(result, index=column.index, name=column.name)

    @staticmethod
    def batch_transform(df: DataFrame, batch_size: int, transform_fn: Callable, name: str = None) -> DataFrame:
        name = name or "Batch Transform"
        batches = to_batches(df, batch_size)
        transform = transform_fn()
        out_batches = [transform(batch.reset_index(drop=True)) for batch in tqdm(batches, desc=name)]
        out_df = from_batches(out_batches).reset_index(drop=True)
        return out_df


class LocalTrainingMixin:
    @staticmethod
    def initialize():
        init_dist_strategy("local")

    @staticmethod
    def initialize_pytorch(*args, **kwargs):
        initialize_pytorch(*args, **kwargs)

    def create_trainer(self, config: BaseTrainerConfig, model: BaseModel, **kwargs) -> BaseTrainer:
        from ludwig.trainers.registry import get_llm_trainers_registry, get_trainers_registry

        if model.type() == MODEL_LLM:
            trainer_cls = get_from_registry(config.type, get_llm_trainers_registry())
        else:
            trainer_cls = get_from_registry(model.type(), get_trainers_registry())

        return trainer_cls(config=config, model=model, **kwargs)

    @staticmethod
    def create_predictor(model: BaseModel, **kwargs):
        from ludwig.models.predictor import get_predictor_cls

        return get_predictor_cls(model.type())(model, **kwargs)

    def sync_model(self, model):
        pass

    @staticmethod
    def broadcast_return(fn):
        return fn()

    @staticmethod
    def is_coordinator() -> bool:
        return True

    @staticmethod
    def tune_batch_size(evaluator_cls: type[BatchSizeEvaluator], dataset_len: int) -> int:
        evaluator = evaluator_cls()
        return evaluator.select_best_batch_size(dataset_len)


class RemoteTrainingMixin:
    def sync_model(self, model):
        pass

    @staticmethod
    def broadcast_return(fn):
        return fn()

    @staticmethod
    def is_coordinator() -> bool:
        return True


@DeveloperAPI
class LocalBackend(LocalPreprocessingMixin, LocalTrainingMixin, Backend):
    BACKEND_TYPE = "local"

    @classmethod
    def shared_instance(cls):
        """Returns a shared singleton LocalBackend instance."""
        if not hasattr(cls, "_shared_instance"):
            cls._shared_instance = cls()
        return cls._shared_instance

    def __init__(self, **kwargs):
        super().__init__(dataset_manager=PandasDatasetManager(self), **kwargs)

    @property
    def num_nodes(self) -> int:
        return 1

    @property
    def num_training_workers(self) -> int:
        return 1

    def get_available_resources(self) -> Resources:
        return Resources(cpus=psutil.cpu_count(), gpus=torch.cuda.device_count())

    def max_concurrent_trials(self, hyperopt_config: HyperoptConfigDict) -> int | None:
        # Every trial will be run with Pandas and NO Ray Datasets. Allow Ray Tune to use all the
        # trial resources it wants, because there is no Ray Datasets process to compete with it for CPUs.
        return None


@DeveloperAPI
class DataParallelBackend(LocalPreprocessingMixin, Backend, ABC):
    BACKEND_TYPE = "deepspeed"

    def __init__(self, **kwargs):
        super().__init__(dataset_manager=PandasDatasetManager(self), **kwargs)
        self._distributed: DistributedStrategy | None = None

    @abstractmethod
    def initialize(self):
        pass

    def initialize_pytorch(self, *args, **kwargs):
        initialize_pytorch(
            *args, local_rank=self._distributed.local_rank(), local_size=self._distributed.local_size(), **kwargs
        )

    def create_trainer(self, **kwargs) -> BaseTrainer:
        from ludwig.trainers.trainer import Trainer

        return Trainer(distributed=self._distributed, **kwargs)

    def create_predictor(self, model: BaseModel, **kwargs):
        from ludwig.models.predictor import get_predictor_cls

        return get_predictor_cls(model.type())(model, distributed=self._distributed, **kwargs)

    def sync_model(self, model):
        # Model weights are only saved on the coordinator, so broadcast
        # to all other ranks
        self._distributed.sync_model(model)

    def broadcast_return(self, fn):
        """Returns the result of calling `fn` on coordinator, broadcast to all other ranks.

        Specifically, `fn` is only executed on coordinator, but its result is returned by every rank by broadcasting the
        return value from coordinator.
        """
        result = fn() if self.is_coordinator() else None
        if self._distributed:
            name = f"broadcast_return_{int(time.time())}"
            result = self._distributed.broadcast_object(result, name=name)
        return result

    def is_coordinator(self):
        return self._distributed.rank() == 0

    @property
    def num_nodes(self) -> int:
        return self._distributed.size() // self._distributed.local_size()

    @property
    def num_training_workers(self) -> int:
        return self._distributed.size()

    def get_available_resources(self) -> Resources:
        # TODO(travis): this double-counts on the same device, it should use a cross-communicator instead
        cpus = torch.as_tensor([psutil.cpu_count()], dtype=torch.int)
        cpus = self._distributed.allreduce(cpus).item()

        gpus = torch.as_tensor([torch.cuda.device_count()], dtype=torch.int)
        gpus = self._distributed.allreduce(gpus).item()

        return Resources(cpus=cpus, gpus=gpus)

    def max_concurrent_trials(self, hyperopt_config: HyperoptConfigDict) -> int | None:
        # Return None since there is no Ray component
        return None

    def tune_batch_size(self, evaluator_cls: type[BatchSizeEvaluator], dataset_len: int) -> int:
        evaluator = evaluator_cls()
        return evaluator.select_best_batch_size(dataset_len)



from typing import List

import torch
from torch import Tensor


def extract_generated_tokens(
    raw_generated_output_sequences: List[Tensor],
    input_lengths: List[int],
    max_new_tokens: int,
    pad_sequence: bool,
) -> List[Tensor]:
    """Extracts the generated tokens from the raw output sequences of the language model.

    Args:
        raw_generated_output_sequences: The raw output sequences of the language model.
            Represented as a list to handle variable length sequences.
        input_lengths: The length of the inputs to the language model.
        max_new_tokens: The maximum number of new tokens that were generated. Used to
            pad the generated sequences to the max_new_tokens.
        pad_sequence: Whether to pad the generated sequences to the max_new_tokens.

    Returns:
        The generated tokens.
    """
    if len(raw_generated_output_sequences) != len(input_lengths):
        raise ValueError(
            f"The number of raw_generated_output_sequences ({len(raw_generated_output_sequences)}) "
            f"must be the same as the number of input_lengths ({len(input_lengths)})."
        )

    generated_outputs = []
    for idx, input_length in enumerate(input_lengths):
        # Remove the input sequence from the generated sequence
        generated_sequence = raw_generated_output_sequences[idx][input_length:]

        # Pad the sequence if it is shorter than the max_new_tokens for downstream metric computation
        if pad_sequence and generated_sequence.size()[0] < max_new_tokens:
            generated_sequence = torch.nn.functional.pad(
                generated_sequence, (0, max_new_tokens - generated_sequence.size()[0]), "constant", 0
            )
        generated_outputs.append(generated_sequence)
    return generated_outputs



import datetime
import time

import pandas as pd
import pytest
from dateutil.parser import parse

from ludwig.api import LudwigModel
from ludwig.constants import (
    BACKEND,
    BINARY,
    DATE,
    EPOCHS,
    FILL_WITH_CONST,
    INPUT_FEATURES,
    MISSING_VALUE_STRATEGY,
    NAME,
    OUTPUT_FEATURES,
    PREPROCESSING,
    RAY,
    TRAINER,
    TYPE,
)
from ludwig.utils.date_utils import create_vector_from_datetime_obj

ray = pytest.importorskip("ray")

pytestmark = [
    pytest.mark.distributed,
]


@pytest.fixture(scope="module")
def string_date_df() -> "pd.DataFrame":
    df = pd.DataFrame.from_dict(
        {
            "date_feature": [str(datetime.datetime.now()) for i in range(100)],
            "binary_feature": [i % 2 for i in range(100)],
        }
    )
    return df


@pytest.fixture(scope="module")
def int_date_df() -> "pd.DataFrame":
    df = pd.DataFrame.from_dict(
        {
            "date_feature": [time.time_ns() for i in range(100)],
            "binary_feature": [i % 2 for i in range(100)],
        }
    )
    return df


@pytest.fixture(scope="module")
def float_date_df() -> "pd.DataFrame":
    df = pd.DataFrame.from_dict(
        {
            "date_feature": [time.time() for i in range(100)],
            "binary_feature": [i % 2 for i in range(100)],
        }
    )
    return df


@pytest.mark.parametrize(
    "date_df",
    [
        pytest.param("string_date_df", id="string_date"),
        pytest.param("int_date_df", id="int_date"),
        pytest.param("float_date_df", id="float_date"),
    ],
)
def test_date_feature_formats(date_df, request, ray_cluster_2cpu):
    df = request.getfixturevalue(date_df)

    config = {
        INPUT_FEATURES: [
            {
                NAME: "date_feature",
                TYPE: DATE,
                PREPROCESSING: {MISSING_VALUE_STRATEGY: FILL_WITH_CONST, "fill_value": "1970-01-01 00:00:00"},
            }
        ],
        OUTPUT_FEATURES: [{NAME: "binary_feature", TYPE: BINARY}],
        TRAINER: {EPOCHS: 2},
        BACKEND: {TYPE: RAY, "processor": {TYPE: "dask"}},
    }

    fill_value = create_vector_from_datetime_obj(parse("1970-01-01 00:00:00"))

    model = LudwigModel(config)
    preprocessed = model.preprocess(df)

    # Because parsing errors are suppressed, we want to ensure that the data was preprocessed correctly. Sample data is
    # drawn from the current time, so the recorded years should not match the fill value's year.
    for date in preprocessed.training_set.to_df().compute().iloc[:, 0].values:
        assert date[0] != fill_value[0]

    for date in preprocessed.validation_set.to_df().compute().iloc[:, 0].values:
        assert date[0] != fill_value[0]

    for date in preprocessed.test_set.to_df().compute().iloc[:, 0].values:
        assert date[0] != fill_value[0]
